
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sm_general_solve.c                                     */
/*  DESCRIPTION:  KheGeneralSolve2018().                                     */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "howard_a.h"
#include "howard_n.h"

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0
#define DEBUG4 0
/* ***
#define DEBUG4 1		** malloc_stats **
#include <malloc.h>
*** */

/*****************************************************************************/
/*                                                                           */
/*  void KheDebugStage(KHE_SOLN soln, KHE_OPTIONS options,                   */
/*    char *stage, int verbosity)                                            */
/*                                                                           */
/*  Print a debug message showing the stage, how much time has been spent    */
/*  so far, and the solution cost.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDebugStage(KHE_SOLN soln, KHE_OPTIONS options,
  char *stage, int verbosity)
{
  KHE_TIMER timer;
  float secs ** , kbytes **;  char time_buff[100] ** , mem_buff[100] **
  HnAssert(soln != NULL, "KheDebugStage internal error");

  ** sort out time usage **
  if( !KheOptionsContainsTimer(options, "global", &timer) )
    sprintf(time_buff, "no time limit");
  else
  {
    secs = KheTimerElapsedTime(timer);
    if( secs > 300.0 )
      sprintf(time_buff, "%.2f mins", secs / 60.0);
    else
      sprintf(time_buff, "%.2f secs", secs);
  }

  ** sort out memory usage **
  ** ***
  kbytes = KheStatsMemory();
  if( kbytes > 1024.0 )
    sprintf(mem_buff, "%.2fMB", kbytes / 1024.0);
  else
    sprintf(mem_buff, "%.2fKB", kbytes);
  *** **

  ** print the header and soln stuff **
  fprintf(stderr, "  KheGeneralSolve2018 %s (%s):\n", stage, time_buff);
  KheSolnDebug(soln, verbosity, 2, stderr);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheInstanceContainsSoftSplitConstraint(KHE_INSTANCE ins)            */
/*                                                                           */
/*  Return true if ins contains at least one soft split events or            */
/*  distribute split events constraint.                                      */
/*                                                                           */
/*****************************************************************************/

static bool KheInstanceContainsSoftSplitConstraint(KHE_INSTANCE ins)
{
  int i;  KHE_CONSTRAINT c;
  for( i = 0;  i < KheInstanceConstraintCount(ins);  i++ )
  {
    c = KheInstanceConstraint(ins, i);
    if( (KheConstraintTag(c) == KHE_SPLIT_EVENTS_CONSTRAINT_TAG ||
        KheConstraintTag(c) == KHE_DISTRIBUTE_SPLIT_EVENTS_CONSTRAINT_TAG)
	&& !KheConstraintRequired(c) && KheConstraintWeight(c) > 0 )
      return true;
  }
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  float KheConvertTimeLimit(char *str)                                     */
/*                                                                           */
/*  Convert *str into a number of seconds, taking the format to be           */
/*                                                                           */
/*      secs    or    mins:secs    or    hrs:mins:secs                       */
/*                                                                           */
/*****************************************************************************/

/* ** moved to KheTimeFromString in khe_sm_timer.c
float KheConve rtTimeLimit(char *str)
{
  char *p;  int hrs, mins, secs;

  ** "-" special case **
  if( strcmp(str, "-") == 0 )
    return -1;

  ** secs **
  p = strstr(str, ":");
  if( p == NULL )
  {
    if( sscanf(str, "%d", &secs) != 1 )
      HnAbort("KheConvertTimeLimit:  format error in time limit %s", str);
    return (float) secs;
  }

  ** mins:secs **
  p = strstr(p + 1, ":");
  if( p == NULL )
  {
    if( sscanf(str, "%d:%d", &mins, &secs) != 2 )
      HnAbort("KheConvertTimeLimit:  format error in time limit %s", str);
    return (float) (mins * 60 + secs);
  }

  ** hrs:mins:secs **
  if( sscanf(str, "%d:%d:%d", &hrs, &mins, &secs) != 3 )
    HnAbort("KheConvertTimeLimit:  format error in time limit %s", str);
  return (float) (hrs * 60 * 60 + mins * 60 + secs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLN KheGeneralSolve2018(KHE_SOLN soln, KHE_OPTIONS options)         */
/*                                                                           */
/*  Solve soln, assuming that it has just emerged from KheSolnMake.          */
/*                                                                           */
/****************************************************************************/
/* extern void KheEjectionChainNodeGroupForTimingTest(KHE_NODE parent_node); */

KHE_SOLN KheGeneralSolve2018(KHE_SOLN soln, KHE_OPTIONS options)
{
  int i;  KHE_EVENT junk;  KHE_NODE cycle_node;  KHE_TASKING tasking;
  char *monitor_id;  KHE_MONITOR m;  KHE_TIMER global_timer;
  KHE_EVENT_TIMETABLE_MONITOR etm;  KHE_INSTANCE ins;  int diversifier;
  bool matching_off, monitor_evenness, time_assignment_only, unassign_off;
  float time_limit;


  /***************************************************************************/
  /*                                                                         */
  /*  initialization phase                                                   */
  /*                                                                         */
  /***************************************************************************/

  /* set soft time limit */
  time_limit = KheTimeFromString(KheOptionsGet(options, "gs_time_limit", "-"));
  global_timer = KheOptionsAddTimer(options, "global", time_limit);
  /* ***
  time_limit = (strcmp(gs_time_limit, "-") == 0 ? KHE_NO_TIME :
    KheConvertTimeLimit(gs_time_limit));
  KheOptionsTimeLimitBegin(options, "global", time_limit);
    KheOptionsSetTimeLimit(options, KheConvertTimeLimit(gs_time_limit));
  *** */

  ins = KheSolnInstance(soln);
  matching_off = KheOptionsGetBool(options, "gs_matching_off", false);
  monitor_evenness = KheOptionsGetBool(options, "gs_monitor_evenness", false);
  time_assignment_only = KheOptionsGetBool(options, "gs_time_assignment_only",
    false);
  unassign_off = KheOptionsGetBool(options, "gs_unassignment_off", false);
  diversifier = KheOptionsGetInt(options, "gs_diversifier", -1);
  if( diversifier != -1 )
    KheSolnSetDiversifier(soln, diversifier);
  if( DEBUG1 )
  {
    fprintf(stderr, "[ KheGeneralSolve2018(%p %s) div %d\n", (void *) soln,
      KheInstanceId(ins) == NULL ? "-" : KheInstanceId(ins),
      KheSolnDiversifier(soln));
    /* KheDebugStage(soln, options, "at start", 2); */
    KheSolveDebug(soln, options, "at start");
  }

  /* set debug_monitor option if debug_monitor_id is present */
  monitor_id = KheOptionsGet(options, "gs_debug_monitor_id", NULL);
  if( monitor_id != NULL )
  {
    if( DEBUG3 )
      fprintf(stderr, "  KheGeneralSolve2018 found debug_monitor_id %s\n",
	monitor_id);
    if( KheSolnRetrieveMonitor(soln, monitor_id, &m) )
    {
      if( DEBUG3 )
      {
	fprintf(stderr, "  KheGeneralSolve2018 setting debug_monitor:\n");
	KheMonitorDebug(m, 1, 2, stderr);
      }
      KheOptionsSetObject(options, "gs_debug_monitor", (void *) m);
    }
    else
      HnAbort("KheGeneralSolve2018: cannot find gs_debug_monitor_id \"%s\"\n",
	monitor_id);
  }


  /***************************************************************************/
  /*                                                                         */
  /*  structural phase                                                       */
  /*                                                                         */
  /***************************************************************************/

  /* split initial cycle meet and make complete representation */
  KheSolnSplitCycleMeet(soln);
  if( !KheSolnMakeCompleteRepresentation(soln, &junk) )
    HnAbort("KheGeneralSolve2018:  !KheSolnMakeCompleteRepresentation");
  if( DEBUG1 )
    KheSolveDebug(soln, options, "after KheSolnMakeCompleteRepresentation");

  /* build layer tree */
  cycle_node = KheLayerTreeMake(soln);
  if( DEBUG2 )
  {
    fprintf(stderr, "  cycle node after KheLayerTreeMake:\n");
    KheNodeDebug(cycle_node, 4, 2, stderr);
  }
  KheOptionsSetBool(options, "es_split_moves",
    KheInstanceContainsSoftSplitConstraint(ins));
  /* ***
  KheOptionsSetEjectorUseSplitMoves(options,
    KheInstanceContainsSoftSplitConstraint(ins));
  *** */

  /* optionally install the global tixel matching */
  if( !matching_off )
  {
    KheSolnMatchingBegin(soln);
    KheSolnMatchingSetWeight(soln, KheCost(1, 0));
    KheSolnMatchingAddAllWorkloadRequirements(soln);
    KheSolnMatchingAttachAllOrdinaryDemandMonitors(soln);
    if( DEBUG2 )
    {
      /* KheDebugStage(soln, options, "after installing the matching", 2); */
      KheSolveDebug(soln, options, "after installing the matching");
      if( KheSolnMatchingDefectCount(soln) > 0 )
	KheGroupMonitorDefectDebug((KHE_GROUP_MONITOR) soln, 2, 2, stderr);
    }
  }
  if( DEBUG1 )
    KheSolveDebug(soln, options, "after KheSolnMatchingBegin");

  /* optionally install evenness monitors */
  if( monitor_evenness )
  {
    KheSolnEvennessBegin(soln);
    KheSolnSetAllEvennessMonitorWeights(soln, KheCost(0, 5));
    KheSolnAttachAllEvennessMonitors(soln);
  }

  /* build task tree, including assigning preassigned resources */
  KheOptionsSetBool(options, "rs_invariant", true);
  KheTaskTreeMake(soln, options);
  /* ***
  KheTaskTreeMake(soln, KHE_TASK_JOB_HARD_PRC | KHE_TASK_JOB_HARD_ASAC |
    KHE_TASK_JOB_SOFT_ASAC, options);
  *** */

  /* find time-equivalent events and resources */
  /* done on first use now KheTimeEquivSolve(time_eq uiv, soln); */

  /* coordinate layers and build runarounds */
  KheCoordinateLayers(cycle_node, true);
  if( DEBUG1 )
    KheSolveDebug(soln, options, "after KheCoordinateLayers");
  if( DEBUG2 )
  {
    fprintf(stderr, "  cycle node after KheCoordinateLayers:\n");
    KheNodeDebug(cycle_node, 4, 2, stderr);
  }
  KheBuildRunarounds(cycle_node, &KheNodeSimpleAssignTimes, options,
    &KheRunaroundNodeAssignTimes, options);
  if( DEBUG1 )
    KheSolveDebug(soln, options, "after KheBuildRunarounds");
  if( DEBUG2 )
  {
    fprintf(stderr, "  cycle node after KheBuildRunarounds:\n");
    KheNodeDebug(cycle_node, 4, 2, stderr);
  }


  /***************************************************************************/
  /*                                                                         */
  /*  time assignment phase                                                  */
  /*                                                                         */
  /***************************************************************************/

  /* assign to nodes below the cycle node */
  for( i = 0;  i < KheNodeChildCount(cycle_node);  i++ )
    KheNodeRecursiveAssignTimes(KheNodeChild(cycle_node, i),
      &KheRunaroundNodeAssignTimes, options);
  if( DEBUG1 )
    KheSolveDebug(soln, options, "after %d KheNodeRecursiveAssignTimes",
      KheNodeChildCount(cycle_node));

  /* ***
  if( DEBUG1 )
    KheDebugStage(soln, options, "before grouping timing test", 2);
  for( i = 0;  i < 10000;  i++ )
    KheEjectionChainNodeGroupForTimingTest(cycle_node);
  if( DEBUG1 )
    KheDebugStage(soln, options, "after grouping timing test", 2);
  *** */

  /* assign to the cycle node */
  /* ***
  if( DEBUG1 )
    KheDebugStage(soln, options, "before time assignment", 2);
  *** */
  KheCycleNodeAssignTimes(cycle_node, options);
  if( DEBUG1 )
    KheSolveDebug(soln, options, "after time assignment");
    /* KheDebugStage(soln, options, "after time assignment", 2); */
  if( time_assignment_only )
  {
    KheSolnSetRunningTime(soln, KheTimerElapsedTime(global_timer));
    KheOptionsDeleteTimer(options, global_timer);
    /* KheSolnSetRunningTime(soln, KheOptionsTimeLimitEnd(options,"global")); */
    /* KheSolnSetRunningTime(soln, KheOptionsTimeNow(options)); */
    if( DEBUG1 )
      fprintf(stderr, "] KheGeneralSolve2018 returning after time asst\n");
    return soln;
  }


  /***************************************************************************/
  /*                                                                         */
  /*  resource assignment phases                                             */
  /*                                                                         */
  /***************************************************************************/

  /* end the inclusion of demand cost in the solution cost */
  KheDisconnectAllDemandMonitors(soln, NULL);

  /* add an event timetable monitor to options */
  etm = KheEventTimetableMonitorMake(soln, KheInstanceFullEventGroup(ins));
  KheMonitorAttachToSoln((KHE_MONITOR) etm);
  KheOptionsSetObject(options, "gs_event_timetable_monitor", (void *) etm);

  /* assign taskings (parts 1 and 2) */
  for( i = 0;  i < KheSolnTaskingCount(soln);  i++ )
  {
    tasking = KheSolnTasking(soln, i);
    KheTaskingAssignResourcesStage1(tasking, options);
    KheTaskingAssignResourcesStage2(tasking, options);
    if( DEBUG1 )
    {
      KheSolveDebug(soln, options, "after %s assignment (parts 1 and 2)",
	KheResourceTypeName(KheTaskingResourceType(tasking)));
      /* ***
      sprintf(buff, "after %s assignment (parts 1 and 2)",
	KheResourceTypeName(KheTaskingResourceType(tasking)));
      KheDebugStage(soln, options, buff, 2);
      *** */
    }
  }

  /* assign taskings (part 3) */
  for( i = 0;  i < KheSolnTaskingCount(soln);  i++ )
  {
    tasking = KheSolnTasking(soln, i);
    KheTaskingAssignResourcesStage3(tasking, options);
    if( DEBUG1 )
    {
      KheSolveDebug(soln, options, "after %s assignment (part 3)",
	KheResourceTypeName(KheTaskingResourceType(tasking)));
      /* ***
      sprintf(buff, "after %s assignment (part 3)",
	KheResourceTypeName(KheTaskingResourceType(tasking)));
      KheDebugStage(soln, options, buff, 2);
      *** */
    }
  }

  /* remove the event timetable monitor from options and delete it */
  HnAssert(KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL)
    == (void *) etm, "KheGeneralSolve2018 internal error (etm)");
  KheOptionsSetObject(options, "gs_event_timetable_monitor", NULL);
  /* KheEventTimetableMonitorDelete(etm); */

  /* try a tree search wherever there are still problems */
  /* ***
  KheTreeSearchRepairTimes(soln, NULL, true);
  *** */


  /***************************************************************************/
  /*                                                                         */
  /*  cleanup phase                                                          */
  /*                                                                         */
  /***************************************************************************/

  /* ensure that the solution cost is the official cost */
  KheSolnEnsureOfficialCost(soln);

  /* merge meets (requires split events monitors, hence this placement) */
  KheMergeMeets(soln);

  /* try task and meet unassignments */
  if( !unassign_off )
  {
    KheSolnTryTaskUnAssignments(soln);
    KheSolnTryMeetUnAssignments(soln);
  }

  /* wrapup */
  if( DEBUG1 )
  {
    KheSolveDebug(soln, options, "at end");
    /* KheDebugStage(soln, options, "at end", 2); */
    if( DEBUG4 )
    {
      KHE_RESOURCE_TIMETABLE_MONITOR rtm;  KHE_RESOURCE r;
      for( i = 0;  i < KheInstanceResourceCount(ins);  i++ )
      {
	r = KheInstanceResource(ins, i);
	fprintf(stderr, "\n  Timetable for resource %s\n", KheResourceId(r));
	rtm = KheResourceTimetableMonitor(soln, r);
	KheResourceTimetableMonitorPrintTimetable(rtm, 6, 2, stderr);
      }
    }
    fprintf(stderr, "] KheGeneralSolve2018 returning\n");
  }
  KheSolnSetRunningTime(soln, KheTimerElapsedTime(global_timer));
  KheOptionsDeleteTimer(options, global_timer);
  /* KheOptionsTimerSetSaveLog(options, soln); */
  /* KheSolnSetRunningTime(soln, KheOptionsTimeLimitEnd(options, "global")); */
  /* KheSolnSetRunningTime(soln, KheOptionsTimeNow(options)); */
  /* ***
  if( KHE_USE_MEMORY_MONITORING && DEBUG4 )
    malloc_stats();
  *** */
  return soln;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSolveDebug(KHE_SOLN soln, KHE_OPTIONS options, int indent,       */
/*    FILE *fp, char *fmt, ...)                                              */
/*                                                                           */
/*  Debug print of the solve.                                                */
/*                                                                           */
/*****************************************************************************/

void KheSolveDebug(KHE_SOLN soln, KHE_OPTIONS options, char *fmt, ...)
{
  KHE_INSTANCE ins;  KHE_TIMER timer;  char buff[20];  va_list args;
  KHE_MONITOR m;
  ins = KheSolnInstance(soln);
  fprintf(stderr, "%*s%s#%d: %.5f", 2, "", KheInstanceId(ins),
    KheSolnDiversifier(soln), KheCostShow(KheSolnCost(soln)));
  if( KheOptionsContainsTimer(options, "global", &timer) )
    fprintf(stderr, ", %s", KheTimeShow(KheTimerElapsedTime(timer), buff));
  if( fmt != NULL )
  {
    fprintf(stderr, " ");
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);
  }
  fprintf(stderr, "\n");
  m = (KHE_MONITOR) KheOptionsGetObject(options, "gs_debug_monitor", NULL);
  if( m != NULL )
    KheMonitorDebug(m, 2, 4, stderr);
}
