
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe.h                                                      */
/*  DESCRIPTION:  Include this file whenever you use KHE                     */
/*                                                                           */
/*  This file has two parts:  type declarations and function declarations.   */
/*  Each part is organized to parallel the User's Guide exactly:             */
/*                                                                           */
/*                    Part B:  Solvers                                       */
/*                                                                           */
/*    Chapter 8.   Introducing Solvers                                       */
/*    Chapter 9.   Time-Structural Solvers                                   */
/*    Chapter 10.  Time Solvers                                              */
/*    Chapter 11.  Resource-Structural Solvers                               */
/*    Chapter 12.  Resource Solvers                                          */
/*    Chapter 13.  Ejection Chains                                           */
/*                                                                           */
/*  This makes it easy to verify that KHE offers what the User's Guide says  */
/*  it offers.                                                               */
/*                                                                           */
/*****************************************************************************/
#ifndef KHE_SOLVERS_HEADER_FILE
#define KHE_SOLVERS_HEADER_FILE

#define KHE_USE_PTHREAD 1
#define KHE_USE_TIMING 1
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include "khe_platform.h"
#include "howard_a.h"
#include "howard_n.h"

#define KHE_ANY_DURATION 0
#define KHE_COST_SHOW_DIGITS 99999


/*****************************************************************************/
/*                                                                           */
/*                        TYPE DECLARATIONS                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*                    Part B:  Solvers                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*    Chapter 8.  Introducing Solvers                                        */
/*                                                                           */
/*****************************************************************************/
 
#define KHE_NO_TIME -1.0
typedef struct khe_timer_rec *KHE_TIMER;
typedef struct khe_timer_set_rec *KHE_TIMER_SET;
typedef struct khe_options_rec *KHE_OPTIONS;

typedef KHE_SOLN (*KHE_GENERAL_SOLVER)(KHE_SOLN soln, KHE_OPTIONS options);

typedef enum {
  KHE_STATS_TABLE_PLAIN,
  KHE_STATS_TABLE_LOUT,
  KHE_STATS_TABLE_LATEX
} KHE_STATS_TABLE_TYPE;

typedef enum {
  KHE_STATS_DATASET_HISTO
} KHE_STATS_DATASET_TYPE;

/* typedef struct khe_stats_timer_rec *KHE_STATS_TIMER; */

typedef enum {
  KHE_BACKOFF_NONE,
  KHE_BACKOFF_EXPONENTIAL
} KHE_BACKOFF_TYPE;

typedef struct khe_backoff_rec *KHE_BACKOFF;


/*****************************************************************************/
/*                                                                           */
/*    Chapter 9.   Time-Structural Solvers                                   */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_time_equiv_rec *KHE_TIME_EQUIV;

typedef enum {
  KHE_SUBTAG_SPLIT_EVENTS,	      /* "SplitEventsGroupMonitor"           */
  KHE_SUBTAG_DISTRIBUTE_SPLIT_EVENTS, /* "DistributeSplitEventsGroupMonitor" */
  KHE_SUBTAG_ASSIGN_TIME,	      /* "AssignTimeGroupMonitor"            */
  KHE_SUBTAG_PREFER_TIMES,	      /* "PreferTimesGroupMonitor"           */
  KHE_SUBTAG_SPREAD_EVENTS,	      /* "SpreadEventsGroupMonitor"          */
  KHE_SUBTAG_LINK_EVENTS,	      /* "LinkEventsGroupMonitor"            */
  KHE_SUBTAG_ORDER_EVENTS,	      /* "OrderEventsGroupMonitor"           */
  KHE_SUBTAG_ASSIGN_RESOURCE,	      /* "AssignResourceGroupMonitor"        */
  KHE_SUBTAG_PREFER_RESOURCES,	      /* "PreferResourcesGroupMonitor"       */
  KHE_SUBTAG_AVOID_SPLIT_ASSIGNMENTS, /* "AvoidSplitAssignmentsGroupMonitor" */
  KHE_SUBTAG_AVOID_CLASHES,	      /* "AvoidClashesGroupMonitor"          */
  KHE_SUBTAG_AVOID_UNAVAILABLE_TIMES, /* "AvoidUnavailableTimesGroupMonitor" */
  KHE_SUBTAG_LIMIT_IDLE_TIMES,	      /* "LimitIdleTimesGroupMonitor"        */
  KHE_SUBTAG_CLUSTER_BUSY_TIMES,      /* "ClusterBusyTimesGroupMonitor"      */
  KHE_SUBTAG_LIMIT_BUSY_TIMES,	      /* "LimitBusyTimesGroupMonitor"        */
  KHE_SUBTAG_LIMIT_WORKLOAD,	      /* "LimitWorkloadGroupMonitor"         */
  KHE_SUBTAG_LIMIT_ACTIVE_INTERVALS,  /* "LimitActiveIntervalsGroupMonitor"  */
  KHE_SUBTAG_LIMIT_RESOURCES,         /* "LimitResourcesGroupMonitor"        */
  KHE_SUBTAG_ORDINARY_DEMAND,	      /* "OrdinaryDemandGroupMonitor"        */
  KHE_SUBTAG_WORKLOAD_DEMAND,	      /* "WorkloadDemandGroupMonitor"        */
  KHE_SUBTAG_KEMPE_DEMAND,	      /* "KempeDemandGroupMonitor"           */
  KHE_SUBTAG_NODE_TIME_REPAIR,	      /* "NodeTimeRepairGroupMonitor"        */
  KHE_SUBTAG_LAYER_TIME_REPAIR,	      /* "LayerTimeRepairGroupMonitor"       */
  KHE_SUBTAG_TASKING,		      /* "TaskingGroupMonitor"               */
  KHE_SUBTAG_RESOURCE_REPAIR,	      /* "ResourceRepairGroupMonitor"        */
  KHE_SUBTAG_ALL_DEMAND		      /* "AllDemandGroupMonitor"             */
} KHE_SUBTAG_STANDARD_TYPE;

typedef struct khe_split_analyser_rec *KHE_SPLIT_ANALYSER;


/*****************************************************************************/
/*                                                                           */
/*    Chapter 10.   Time Solvers                                             */
/*                                                                           */
/*****************************************************************************/

typedef bool (*KHE_NODE_TIME_SOLVER)(KHE_NODE parent_node, KHE_OPTIONS options);
typedef bool (*KHE_LAYER_TIME_SOLVER)(KHE_LAYER layer, KHE_OPTIONS options);

typedef struct khe_kempe_stats_rec *KHE_KEMPE_STATS;
typedef struct khe_meet_bound_group_rec *KHE_MEET_BOUND_GROUP;

typedef struct khe_layer_asst_rec *KHE_LAYER_ASST;

typedef struct khe_meet_set_solver_rec *KHE_MEET_SET_SOLVER;


/*****************************************************************************/
/*                                                                           */
/*    Chapter 11.  Resource-Structural Solvers                               */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_task_bound_group_rec *KHE_TASK_BOUND_GROUP;

typedef struct khe_task_grouper_rec *KHE_TASK_GROUPER;

typedef struct khe_task_group_rec *KHE_TASK_GROUP;
typedef struct khe_task_groups_rec *KHE_TASK_GROUPS;


/*****************************************************************************/
/*                                                                           */
/*    Chapter 12.  Resource Solvers                                          */
/*                                                                           */
/*****************************************************************************/

typedef bool (*KHE_TASKING_SOLVER)(KHE_TASKING tasking, KHE_OPTIONS options);

typedef struct khe_frame_iterator_rec {
  KHE_FRAME		frame;		/* the frame being iterated over     */
  int			count;		/* 1 + KheFrameTimeGroupCount(frame) */
  int			extra;		/* start point along frame           */
  int			curr;		/* current position of iterator      */
  KHE_RESOURCE		resource;	/* resource whose activity is wanted */
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;	/* timetable monitor of resource     */
} *KHE_FRAME_ITERATOR;

typedef struct khe_resource_matching_solver_rec *KHE_RESOURCE_MATCHING_SOLVER;
typedef struct khe_resource_matching_demand_set_rec
  *KHE_RESOURCE_MATCHING_DEMAND_SET;
typedef HA_ARRAY(KHE_RESOURCE_MATCHING_DEMAND_SET)
  ARRAY_KHE_RESOURCE_MATCHING_DEMAND_SET;

typedef enum {
  KHE_MOVE_UNCHECKED,			/* just a basic unchecked move       */
  KHE_MOVE_CHECKED,			/* do nothing if would cause clashes */
  KHE_MOVE_EJECTING,			/* eject clashes                     */
  KHE_MOVE_KEMPE			/* fix clashes using Kempe chains    */
} KHE_MOVE_TYPE;


/*****************************************************************************/
/*                                                                           */
/*    Chapter 13.  Ejection Chains                                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_ejector_rec *KHE_EJECTOR;

typedef struct khe_ejector_major_schedule_rec *KHE_EJECTOR_MAJOR_SCHEDULE;
typedef struct khe_ejector_minor_schedule_rec *KHE_EJECTOR_MINOR_SCHEDULE;

typedef bool (*KHE_EJECTOR_AUGMENT_FN)(KHE_EJECTOR ej, KHE_MONITOR d);


/*****************************************************************************/
/*                                                                           */
/*                     FUNCTION DECLARATIONS                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*                    Part B:  Solvers                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*    Chapter  8.  Introducing Solvers                                       */
/*                                                                           */
/*****************************************************************************/

/* 8.1 Keeping track of running time */
extern float KheTimeFromString(char *str);
extern char *KheTimeShow(float secs, char buff[20]);

extern KHE_TIMER KheTimerMake(char *tag, float limit_in_seconds, HA_ARENA a);
extern KHE_TIMER KheTimerCopy(KHE_TIMER timer, HA_ARENA a);
extern char *KheTimerTag(KHE_TIMER timer);
extern float KheTimerTimeLimit(KHE_TIMER timer);
extern void KheTimerResetStartTime(KHE_TIMER timer);
extern void KheTimerResetTimeLimit(KHE_TIMER timer, float limit_in_seconds);
extern float KheTimerElapsedTime(KHE_TIMER timer);
extern bool KheTimerTimeLimitReached(KHE_TIMER timer);
extern void KheTimerDebug(KHE_TIMER timer, int verbosity, int indent, FILE *fp);

extern KHE_TIMER_SET KheTimerSetMake(HA_ARENA a);
extern KHE_TIMER_SET KheTimerSetCopy(KHE_TIMER_SET timer_set, HA_ARENA a);
extern void KheTimerSetAddTimer(KHE_TIMER_SET timer_set, KHE_TIMER timer);
extern void KheTimerSetDeleteTimer(KHE_TIMER_SET timer_set, KHE_TIMER timer);
extern bool KheTimerSetContainsTimer(KHE_TIMER_SET timer_set, char *tag,
  KHE_TIMER *timer);
extern bool KheTimerSetTimeLimitReached(KHE_TIMER_SET timer_set);
/* extern void KheTimerSetSaveLog(KHE_TIMER_SET ts, KHE_SOLN soln); */
extern void KheTimerSetDebug(KHE_TIMER_SET timer_set, int verbosity,
  int indent, FILE *fp);

/* 8.2 Options, running time, and time limits */
extern KHE_OPTIONS KheOptionsMake(HA_ARENA a);
extern HA_ARENA KheOptionsArena(KHE_OPTIONS options);
/* extern void KheOptionsDelete(KHE_OPTIONS options); */
extern KHE_OPTIONS KheOptionsCopy(KHE_OPTIONS options, HA_ARENA a);

extern void KheOptionsSet(KHE_OPTIONS options, char *key, char *value);
extern char *KheOptionsGet(KHE_OPTIONS options, char *key, char *dft);

extern void KheOptionsSetBool(KHE_OPTIONS options, char *key, bool value);
extern bool KheOptionsGetBool(KHE_OPTIONS options, char *key, bool dft);

extern void KheOptionsSetInt(KHE_OPTIONS options, char *key, int value);
extern int KheOptionsGetInt(KHE_OPTIONS options, char *key, int dft);

extern void KheOptionsSetObject(KHE_OPTIONS options, char *key, void *value);
extern void *KheOptionsGetObject(KHE_OPTIONS options, char *key, void *dft);

extern KHE_TIMER KheOptionsAddTimer(KHE_OPTIONS options, char *tag,
  float limit_in_seconds);
extern void KheOptionsDeleteTimer(KHE_OPTIONS options, KHE_TIMER timer);
extern bool KheOptionsContainsTimer(KHE_OPTIONS options, char *tag,
  KHE_TIMER *timer);
extern bool KheOptionsTimeLimitReached(KHE_OPTIONS options);
/* ***
extern void KheOptionsTimerSetSaveLog(KHE_OPTIONS options, KHE_SOLN soln);
*** */
extern void KheOptionsTimerSetDebug(KHE_OPTIONS options, int verbosity,
  int indent, FILE *fp);

/* ***
extern void KheOptionsTimeLimitBegin(KHE_OPTIONS options, char *tag,
  float limit_in_secs);
extern float KheOptionsTimeLimitEnd(KHE_OPTIONS options, char *tag);
extern bool KheOptionsTimeLimitNow(KHE_OPTIONS options, char *tag,
  float *elapsed_time_in_secs);
extern bool KheOptionsTimeLimitReached(KHE_OPTIONS options);
extern void KheOptionsTimeLimitReset(KHE_OPTIONS options, char *tag,
  float limit_in_secs);
*** */

/* ***
extern float KheOptionsTimeNow(KHE_OPTIONS options);
extern void KheOptionsSetTimeLimit(KHE_OPTIONS options, float limit_in_secs);
extern float KheOptionsTimeLimit(KHE_OPTIONS options);
extern bool KheOptionsTimeLimitReached(KHE_OPTIONS options);
*** */

extern KHE_FRAME KheOptionsFrame(KHE_OPTIONS options, char *key, KHE_SOLN soln);

/* 8.3 General solving */
extern KHE_SOLN KheGeneralSolve2018(KHE_SOLN soln, KHE_OPTIONS options);
extern float KheConvertTimeLimit(char *str);
extern void KheSolveDebug(KHE_SOLN soln, KHE_OPTIONS options, char *fmt, ...);

/* 8.4 Parallel solving */
extern void KheArchiveParallelSolve(KHE_ARCHIVE archive,
  KHE_GENERAL_SOLVER solver, KHE_OPTIONS options, HA_ARENA_SET as,
  KHE_SOLN_GROUP soln_group);
extern KHE_SOLN KheInstanceParallelSolve(KHE_INSTANCE ins,
  KHE_GENERAL_SOLVER solver, KHE_OPTIONS options, HA_ARENA_SET as);

/* Benchmarking - obsolete */
/* ***
extern bool KheBenchmarkTryInstance(KHE_INSTANCE ins);
extern void KheBenchmark(KHE_ARCHIVE archive, KHE_GENERAL_SOLVER solver,
  char *solver_name, char *author_name, char test_label,
  KHE_STATS_TABLE_TYPE table_type);
*** */

/* 8.5 Gathering statistics */
/* ***
extern KHE_STATS_TIMER KheStatsTimerMake(HA_ARENA a);
extern void KheStatsTimerReset(KHE_STATS_TIMER st);
extern float KheStatsTimerNow(KHE_STATS_TIMER st);
** extern void KheStatsTimerDelete(KHE_STATS_TIMER st); **
extern KHE_STATS_TIMER KheStatsTimerCopy(KHE_STATS_TIMER st, HA_ARENA a);
*** */

extern char *KheStatsDateToday(void);

extern void KheStatsFileBegin(char *file_name);
extern void KheStatsFileEnd(char *file_name);

extern void KheStatsTableBegin(char *file_name, KHE_STATS_TABLE_TYPE table_type,
  int col_width, char *corner, bool with_average_row, bool with_total_row,
  bool highlight_cost_minima, bool highlight_time_minima,
  bool highlight_int_minima);
extern void KheStatsTableEnd(char *file_name);
extern void KheStatsCaptionAdd(char *file_name, char *fmt, ...);
extern void KheStatsRowAdd(char *file_name, char *row_label, bool rule_below);
extern void KheStatsColAdd(char *file_name, char *col_label, bool rule_after);

extern void KheStatsEntryAddString(char *file_name, char *row_label,
  char *col_label, char *str);
extern void KheStatsEntryAddCost(char *file_name, char *row_label,
  char *col_label, KHE_COST cost);
extern void KheStatsEntryAddTime(char *file_name, char *row_label,
  char *col_label, float time);
extern void KheStatsEntryAddInt(char *file_name, char *row_label,
  char *col_label, int val);

extern void KheStatsGraphBegin(char *file_name);
extern void KheStatsGraphEnd(char *file_name);
extern void KheStatsGraphSetWidth(char *file_name, float width);
extern void KheStatsGraphSetHeight(char *file_name, float height);
extern void KheStatsGraphSetXMax(char *file_name, float xmax);
extern void KheStatsGraphSetYMax(char *file_name, float ymax);
extern void KheStatsGraphSetAboveCaption(char *file_name, char *val);
extern void KheStatsGraphSetBelowCaption(char *file_name, char *val);
extern void KheStatsGraphSetLeftCaption(char *file_name, char *val);
extern void KheStatsGraphSetRightCaption(char *file_name, char *val);
extern void KheStatsDataSetAdd(char *file_name, char *dataset_label,
  KHE_STATS_DATASET_TYPE dataset_type);
extern void KheStatsPointAdd(char *file_name, char *dataset_label,
  float x, float y);

/* 8.5 Exponential backoff */
extern KHE_BACKOFF KheBackoffBegin(KHE_BACKOFF_TYPE backoff_type, HA_ARENA a);
extern bool KheBackoffAcceptOpportunity(KHE_BACKOFF bk);
extern void KheBackoffResult(KHE_BACKOFF bk, bool success);
/* extern void KheBackoffEnd(KHE_BACKOFF bk); */

extern char *KheBackoffShowNextDecision(KHE_BACKOFF bk);
extern void KheBackoffDebug(KHE_BACKOFF bk, int verbosity, int indent,FILE *fp);
extern void KheBackoffTest(FILE *fp);


/*****************************************************************************/
/*                                                                           */
/*    Chapter 9.   Structural Solvers                                        */
/*                                                                           */
/*****************************************************************************/

/* 9.1 Layer tree construction */
extern KHE_NODE KheLayerTreeMake(KHE_SOLN soln);

/* 9.2 Time-equivalence */
extern KHE_TIME_EQUIV KheTimeEquivMake(KHE_SOLN soln);
extern void KheTimeEquivDelete(KHE_TIME_EQUIV te);
extern void KheTimeEquivSolve(KHE_TIME_EQUIV te);

extern KHE_TIME_EQUIV KheTimeEquivOption(KHE_OPTIONS options,
  char *key, KHE_SOLN soln);

extern int KheTimeEquivEventGroupCount(KHE_TIME_EQUIV te);
extern KHE_EVENT_GROUP KheTimeEquivEventGroup(KHE_TIME_EQUIV te, int i);
extern KHE_EVENT_GROUP KheTimeEquivEventEventGroup(KHE_TIME_EQUIV te,
  KHE_EVENT e);
extern int KheTimeEquivEventEventGroupIndex(KHE_TIME_EQUIV te, KHE_EVENT e);

extern int KheTimeEquivResourceGroupCount(KHE_TIME_EQUIV te);
extern KHE_RESOURCE_GROUP KheTimeEquivResourceGroup(KHE_TIME_EQUIV te, int i);
extern KHE_RESOURCE_GROUP KheTimeEquivResourceResourceGroup(KHE_TIME_EQUIV te,
  KHE_RESOURCE r);
extern int KheTimeEquivResourceResourceGroupIndex(KHE_TIME_EQUIV te,
  KHE_RESOURCE r);

/* 9.3.1 Layer construction */
extern KHE_LAYER KheLayerMakeFromResource(KHE_NODE parent_node,
  KHE_RESOURCE r);
extern void KheNodeChildLayersMake(KHE_NODE parent_node);
extern void KheNodeChildLayersReduce(KHE_NODE parent_node);

/* 9.3.2 Layer coordination */
extern void KheCoordinateLayers(KHE_NODE parent_node, bool with_domination);

/* 9.4 Runarounds */
extern bool KheMinimumRunaroundDuration(KHE_NODE parent_node,
  KHE_NODE_TIME_SOLVER time_solver, KHE_OPTIONS options, int *duration);
extern void KheBuildRunarounds(KHE_NODE parent_node,
  KHE_NODE_TIME_SOLVER mrd_solver, KHE_OPTIONS mrd_options,
  KHE_NODE_TIME_SOLVER runaround_solver, KHE_OPTIONS runaround_options);

/* 9.5.1 Node merging */
extern bool KheNodeMergeCheck(KHE_NODE node1, KHE_NODE node2);
extern bool KheNodeMerge(KHE_NODE node1, KHE_NODE node2, KHE_NODE *res);

/* 9.5.2 Node meet splitting and merging */
extern void KheNodeMeetSplit(KHE_NODE node, bool recursive);
extern void KheNodeMeetMerge(KHE_NODE node, bool recursive);

/* 9.5.3 Node moving */
extern bool KheNodeMoveCheck(KHE_NODE child_node, KHE_NODE parent_node);
extern bool KheNodeMove(KHE_NODE child_node, KHE_NODE parent_node);

/* 9.5.4 Vizier nodes */
extern KHE_NODE KheNodeVizierMake(KHE_NODE parent_node);
extern void KheNodeVizierDelete(KHE_NODE parent_node);

/* 9.5.5 Flattening */
extern void KheNodeBypass(KHE_NODE node);
extern void KheNodeFlatten(KHE_NODE parent_node);

/* 9.6 Adding zones */
extern void KheLayerInstallZonesInParent(KHE_LAYER layer);
extern void KheNodeExtendZones(KHE_NODE node);

/* 9.7.1 Analysing split defects */
extern KHE_SPLIT_ANALYSER KheSplitAnalyserMake(KHE_SOLN soln);
extern void KheSplitAnalyserDelete(KHE_SPLIT_ANALYSER sa);
extern KHE_SPLIT_ANALYSER KheSplitAnalyserOption(KHE_OPTIONS options,
  char *key, KHE_SOLN soln);

extern void KheSplitAnalyserAnalyse(KHE_SPLIT_ANALYSER sa, KHE_EVENT e);
extern int KheSplitAnalyserSplitSuggestionCount(KHE_SPLIT_ANALYSER sa);
extern void KheSplitAnalyserSplitSuggestion(KHE_SPLIT_ANALYSER sa, int i,
  int *merged_durn, int *split1_durn);
extern int KheSplitAnalyserMergeSuggestionCount(KHE_SPLIT_ANALYSER sa);
extern void KheSplitAnalyserMergeSuggestion(KHE_SPLIT_ANALYSER sa, int i,
  int *split1_durn, int *split2_durn);
extern void KheSplitAnalyserDebug(KHE_SPLIT_ANALYSER sa, int verbosity,
  int indent, FILE *fp);

/* 9.7.2 Merging adjacent meets */
extern void KheMergeMeets(KHE_SOLN soln);

/* 9.8 Monitor attachment and grouping */
extern char *KheSubTagLabel(KHE_SUBTAG_STANDARD_TYPE sub_tag);
extern KHE_SUBTAG_STANDARD_TYPE KheSubTagFromTag(KHE_MONITOR_TAG tag);
extern bool KheMonitorHasParent(KHE_MONITOR m, int sub_tag,
  KHE_GROUP_MONITOR *res_gm);
extern void KheMonitorAddSelfOrParent(KHE_MONITOR m, int sub_tag,
  KHE_GROUP_MONITOR gm);
extern void KheMonitorDeleteAllParentsRecursive(KHE_MONITOR m);


/*****************************************************************************/
/*                                                                           */
/*    Chapter 10.   Time Solvers                                             */
/*                                                                           */
/*****************************************************************************/

/* 10.2.1 Node assignment functions */
extern bool KheNodeMeetSwapCheck(KHE_NODE node1, KHE_NODE node2);
extern bool KheNodeMeetSwap(KHE_NODE node1, KHE_NODE node2);
extern bool KheNodeMeetRegularAssignCheck(KHE_NODE node, KHE_NODE sibling_node);
extern bool KheNodeMeetRegularAssign(KHE_NODE node, KHE_NODE sibling_node);
extern void KheNodeMeetUnAssign(KHE_NODE node);

/* 10.2.2 Kempe and ejecting meet moves */
extern KHE_GROUP_MONITOR KheKempeDemandGroupMonitorMake(KHE_SOLN soln);
extern bool KheKempeMeetMove(KHE_MEET meet, KHE_MEET target_meet,
  int offset, bool preserve_regularity, int *demand, bool *basic,
  KHE_KEMPE_STATS kempe_stats);
extern bool KheKempeMeetMoveTime(KHE_MEET meet, KHE_TIME t,
  bool preserve_regularity, int *demand, bool *basic,
  KHE_KEMPE_STATS kempe_stats);

extern KHE_KEMPE_STATS KheKempeStatsMake(HA_ARENA a);
extern void KheKempeStatsDelete(KHE_KEMPE_STATS kempe_stats);
extern KHE_KEMPE_STATS KheKempeStatsOption(KHE_OPTIONS options, char *key);

extern int KheKempeStatsStepHistoMax(KHE_KEMPE_STATS kempe_stats);
extern int KheKempeStatsStepHistoFrequency(KHE_KEMPE_STATS kempe_stats,
  int step_count);
extern int KheKempeStatsStepHistoTotal(KHE_KEMPE_STATS kempe_stats);
extern float KheKempeStatsStepHistoAverage(KHE_KEMPE_STATS kempe_stats);

extern int KheKempeStatsPhaseHistoMax(KHE_KEMPE_STATS kempe_stats);
extern int KheKempeStatsPhaseHistoFrequency(KHE_KEMPE_STATS kempe_stats,
  int phase_count);
extern int KheKempeStatsPhaseHistoTotal(KHE_KEMPE_STATS kempe_stats);
extern float KheKempeStatsPhaseHistoAverage(KHE_KEMPE_STATS kempe_stats);

extern bool KheEjectingMeetMove(KHE_MEET meet, KHE_MEET target_meet, int offset,
  bool allow_eject, bool preserve_regularity, int *demand, bool *basic);
extern bool KheEjectingMeetMoveTime(KHE_MEET meet, KHE_TIME t,
  bool allow_eject, bool preserve_regularity, int *demand, bool *basic);
extern bool KheBasicMeetMove(KHE_MEET meet, KHE_MEET target_meet,
  int offset, bool preserve_regularity, int *demand);
extern bool KheBasicMeetMoveTime(KHE_MEET meet, KHE_TIME t,
  bool preserve_regularity, int *demand);
 
extern bool KheTypedMeetMove(KHE_MEET meet, KHE_MEET target_meet, int offset,
  KHE_MOVE_TYPE mt, bool preserve_regularity, int *demand, bool *basic,
  KHE_KEMPE_STATS kempe_stats);
extern bool KheTypedMeetMoveTime(KHE_MEET meet, KHE_TIME t,
  KHE_MOVE_TYPE mt, bool preserve_regularity, int *demand, bool *basic,
  KHE_KEMPE_STATS kempe_stats);

/* 10.3.1 Meet bound groups */
extern KHE_MEET_BOUND_GROUP KheMeetBoundGroupMake(KHE_SOLN soln);
extern void KheMeetBoundGroupAddMeetBound(KHE_MEET_BOUND_GROUP mbg,
  KHE_MEET_BOUND mb);
extern int KheMeetBoundGroupMeetBoundCount(KHE_MEET_BOUND_GROUP mbg);
extern KHE_MEET_BOUND KheMeetBoundGroupMeetBound(KHE_MEET_BOUND_GROUP mbg,
  int i);
extern bool KheMeetBoundGroupDelete(KHE_MEET_BOUND_GROUP mbg);

/* 10.3.2 Exposing resource unavailability */
extern void KheMeetAddUnavailableBound(KHE_MEET meet, KHE_COST min_weight,
  KHE_MEET_BOUND_GROUP mbg);
extern void KheSolnAddUnavailableBounds(KHE_SOLN soln, KHE_COST min_weight,
  KHE_MEET_BOUND_GROUP mbg);

/* 10.3.3 Preventing cluster busy times and limit idle times defects */
extern void KheSolnClusterAndLimitMeetDomains(KHE_SOLN soln,
  KHE_COST min_cluster_weight, KHE_COST min_idle_weight,
  float slack, KHE_MEET_BOUND_GROUP mbg, KHE_OPTIONS options);

/* 10.4 Some basic time solvers */
extern bool KheNodeSimpleAssignTimes(KHE_NODE parent_node, KHE_OPTIONS options);
extern bool KheLayerSimpleAssignTimes(KHE_LAYER layer, KHE_OPTIONS options);

extern bool KheNodeRecursiveAssignTimes(KHE_NODE parent_node,
  KHE_NODE_TIME_SOLVER solver, KHE_OPTIONS options);

extern bool KheNodeUnAssignTimes(KHE_NODE parent_node, KHE_OPTIONS options);
extern bool KheLayerUnAssignTimes(KHE_LAYER layer, KHE_OPTIONS options);

extern bool KheNodeAllChildMeetsAssigned(KHE_NODE parent_node);
extern bool KheLayerAllChildMeetsAssigned(KHE_LAYER layer);

extern bool KheNodePreassignedAssignTimes(KHE_NODE root_node,
  KHE_OPTIONS options);
extern bool KheLayerPreassignedAssignTimes(KHE_LAYER layer,
  KHE_OPTIONS options);

extern bool KheSolnTryMeetUnAssignments(KHE_SOLN soln);

/* 10.5 A time solver for runarounds */
extern bool KheRunaroundNodeAssignTimes(KHE_NODE parent_node,
  KHE_OPTIONS options);

/* 10.6 Extended layer matching with Elm */
extern bool KheElmLayerAssign(KHE_LAYER layer,
  KHE_SPREAD_EVENTS_CONSTRAINT sec, KHE_OPTIONS options);

/* 10.7.1 Node-regular time repair using layer node matching */
extern bool KheLayerNodeMatchingNodeRepairTimes(KHE_NODE parent_node,
  KHE_OPTIONS options);
extern bool KheLayerNodeMatchingLayerRepairTimes(KHE_LAYER layer,
  KHE_OPTIONS options);

/* 10.7.3 Tree search layer time repair */
extern bool KheTreeSearchLayerRepairTimes(KHE_SOLN soln, KHE_RESOURCE r);
extern bool KheTreeSearchRepairTimes(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,
  bool with_defects);

/* 10.7.4 Meet set time repair */
extern KHE_MEET_SET_SOLVER KheMeetSetSolveBegin(KHE_SOLN soln, int max_meets);
extern void KheMeetSetSolveAddMeet(KHE_MEET_SET_SOLVER mss, KHE_MEET meet);
extern bool KheMeetSetSolveEnd(KHE_MEET_SET_SOLVER mss);

extern bool KheFuzzyMeetMove(KHE_MEET meet, KHE_MEET target_meet, int offset,
  int width, int depth, int max_meets);

/* 10.8.1 Layer assignments */
extern KHE_LAYER_ASST KheLayerAsstMake(KHE_SOLN soln);
extern void KheLayerAsstDelete(KHE_LAYER_ASST layer_asst);
extern void KheLayerAsstBegin(KHE_LAYER_ASST layer_asst, KHE_LAYER layer);
extern void KheLayerAsstEnd(KHE_LAYER_ASST layer_asst);
extern void KheLayerAsstUndo(KHE_LAYER_ASST layer_asst);
extern void KheLayerAsstRedo(KHE_LAYER_ASST layer_asst);
extern void KheLayerAsstDebug(KHE_LAYER_ASST layer_asst, int verbosity,
  int indent, FILE *fp);

/* 10.8.2 A solver for layered time assignment */
extern bool KheNodeLayeredAssignTimes(KHE_NODE parent_node,KHE_OPTIONS options);
extern int KheNodeLayeredLayerCmp(const void *t1, const void *t2);
extern bool KheLayerParallelAssignTimes(KHE_LAYER layer, KHE_OPTIONS options);

/* 10.8.3 A complete time solver */
extern bool KheCycleNodeAssignTimes(KHE_NODE cycle_node, KHE_OPTIONS options);


/*****************************************************************************/
/*                                                                           */
/*    Chapter 11.  Resource-Structural Solvers                               */
/*                                                                           */
/*****************************************************************************/

/* 11.1 Task bound groups */
extern KHE_TASK_BOUND_GROUP KheTaskBoundGroupMake(KHE_SOLN soln);
extern void KheTaskBoundGroupAddTaskBound(KHE_TASK_BOUND_GROUP tbg,
  KHE_TASK_BOUND tb);
extern int KheTaskBoundGroupTaskBoundCount(KHE_TASK_BOUND_GROUP tbg);
extern KHE_TASK_BOUND KheTaskBoundGroupTaskBound(KHE_TASK_BOUND_GROUP tbg,
  int i);
extern bool KheTaskBoundGroupDelete(KHE_TASK_BOUND_GROUP tbg);

/* 11.3 Task tree construction */
extern bool KheTaskTreeMake(KHE_SOLN soln, KHE_OPTIONS options);
extern KHE_TASKING KheTaskingMakeFromResourceType(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt);
extern bool KheTaskingMakeTaskTree(KHE_TASKING tasking,
  KHE_TASK_BOUND_GROUP tbg, KHE_OPTIONS options);

/* 11.4 Tighten to partition */
extern bool KheTaskingTightenToPartition(KHE_TASKING tasking,
  KHE_TASK_BOUND_GROUP tbg, KHE_OPTIONS options);

/* 11.5 Grouping by resource constraints */
extern bool KheGroupByResourceConstraints(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_OPTIONS options, KHE_TASK_SET ts);

/* 11.6 Grouping by resource */
extern bool KheTaskingGroupByResource(KHE_TASKING tasking, KHE_OPTIONS options,
  KHE_TASK_SET ts);

/* 11.7 The task grouper */
extern KHE_TASK_GROUPER KheTaskGrouperMake(KHE_RESOURCE_TYPE rt, HA_ARENA a);
extern void KheTaskGrouperClear(KHE_TASK_GROUPER tg);
extern bool KheTaskGrouperAddTask(KHE_TASK_GROUPER tg, KHE_TASK t);
/* ***
extern KHE_TASK KheTaskGrouperLeaderTask(KHE_TASK_GROUPER tg, KHE_RESOURCE r);
*** */
extern void KheTaskGrouperGroup(KHE_TASK_GROUPER tg);
extern void KheTaskGrouperUnGroup(KHE_TASK_GROUPER tg);

/* 11.8 Other task tree solvers */
extern void KheTaskingAllowSplitAssignments(KHE_TASKING tasking,
  bool unassigned_only);
extern void KheTaskingEnlargeDomains(KHE_TASKING tasking, bool unassigned_only);

/* 11.9 Task groups */
extern KHE_TASK_GROUPS KheTaskGroupsMakeFromTasking(KHE_TASKING tasking);
extern void KheTaskGroupsDelete(KHE_TASK_GROUPS task_groups);
extern int KheTaskGroupsTaskGroupCount(KHE_TASK_GROUPS task_groups);
extern KHE_TASK_GROUP KheTaskGroupsTaskGroup(KHE_TASK_GROUPS task_groups,
  int i);
extern int KheTaskGroupTaskCount(KHE_TASK_GROUP task_group);
extern KHE_TASK KheTaskGroupTask(KHE_TASK_GROUP task_group, int i);

extern int KheTaskGroupTotalDuration(KHE_TASK_GROUP task_group);
extern float KheTaskGroupTotalWorkload(KHE_TASK_GROUP task_group);
extern KHE_RESOURCE_GROUP KheTaskGroupDomain(KHE_TASK_GROUP task_group);
extern int KheTaskGroupDecreasingDurationCmp(KHE_TASK_GROUP tg1,
  KHE_TASK_GROUP tg2);

extern int KheTaskGroupUnassignedTaskCount(KHE_TASK_GROUP task_group);
extern bool KheTaskGroupAssignCheck(KHE_TASK_GROUP task_group, KHE_RESOURCE r);
extern bool KheTaskGroupAssign(KHE_TASK_GROUP task_group, KHE_RESOURCE r);
extern void KheTaskGroupUnAssign(KHE_TASK_GROUP task_group, KHE_RESOURCE r);

extern void KheTaskGroupDebug(KHE_TASK_GROUP task_group, int verbosity,
  int indent, FILE *fp);
extern void KheTaskGroupsDebug(KHE_TASK_GROUPS task_groups, int verbosity,
  int indent, FILE *fp);


/*****************************************************************************/
/*                                                                           */
/*    Chapter 12.  Resource Solvers                                          */
/*                                                                           */
/*****************************************************************************/

/* 12.2 The resource assignment invariant */
extern void KheAtomicOperationBegin(KHE_SOLN soln, KHE_MARK *mark,
  int *init_count, bool resource_invariant);
extern bool KheAtomicOperationEnd(KHE_SOLN soln, KHE_MARK *mark,
  int *init_count, bool resource_invariant, bool success);
extern void KheDisconnectAllDemandMonitors(KHE_SOLN soln, KHE_RESOURCE_TYPE rt);

/* 12.3 Ordinary, ejecting, and Kempe task and task set moves */
extern bool KheTaskSetMove(KHE_TASK_SET ts, KHE_RESOURCE r);
extern bool KheEjectingTaskMove(KHE_TASK task, KHE_RESOURCE r,
  bool allow_eject);
extern bool KheEjectingTaskMoveFrame(KHE_TASK task, KHE_RESOURCE r,
  bool allow_eject, KHE_FRAME frame);
extern bool KheEjectingTaskSetMove(KHE_TASK_SET ts, KHE_RESOURCE r,
  bool allow_eject);
extern bool KheEjectingTaskSetMoveFrame(KHE_TASK_SET ts, KHE_RESOURCE r,
  bool allow_eject, KHE_FRAME frame);

extern bool KheKempeTaskMove(KHE_TASK task, KHE_RESOURCE r);
extern bool KheKempeTaskMoveFrame(KHE_TASK task, KHE_RESOURCE r,
  KHE_FRAME frame);
extern bool KheKempeTaskSetMove(KHE_TASK_SET ts, KHE_RESOURCE r);
extern bool KheKempeTaskSetMoveFrame(KHE_TASK_SET ts, KHE_RESOURCE r,
  KHE_FRAME frame);

extern bool KheTypedTaskMove(KHE_TASK task, KHE_RESOURCE r, KHE_MOVE_TYPE mt);
extern bool KheTypedTaskMoveFrame(KHE_TASK task, KHE_RESOURCE r,
  KHE_MOVE_TYPE mt, KHE_FRAME frame);
extern bool KheTypedTaskSetMove(KHE_TASK_SET ts, KHE_RESOURCE r,
  KHE_MOVE_TYPE mt);
extern bool KheTypedTaskSetMoveFrame(KHE_TASK_SET ts, KHE_RESOURCE r,
  KHE_MOVE_TYPE mt, KHE_FRAME frame);
extern char *KheMoveTypeShow(KHE_MOVE_TYPE mt);

/* 12.4 Frame operations for resource solvers */
extern bool KheFrameTaskSet(KHE_FRAME frame, KHE_RESOURCE r, KHE_TASK_SET *ts);

extern void KheFrameSetVisitNum(KHE_FRAME frame, KHE_RESOURCE r, int num);
extern int KheFrameVisitNum(KHE_FRAME frame, KHE_RESOURCE r);
extern bool KheFrameVisited(KHE_FRAME frame, KHE_RESOURCE r, int slack);
extern void KheFrameVisit(KHE_FRAME frame, KHE_RESOURCE r);
extern void KheFrameUnVisit(KHE_FRAME frame, KHE_RESOURCE r);

extern bool KheFrameIsActive(KHE_FRAME frame, KHE_RESOURCE r);
extern bool KheFrameIsInactive(KHE_FRAME frame, KHE_RESOURCE r);

/* ***
extern KHE_FRAME KheFrameActiveAtLeft(KHE_FRAME frame, KHE_RESOURCE r);
extern KHE_FRAME KheFrameInactiveAtLeft(KHE_FRAME frame, KHE_RESOURCE r);
extern KHE_FRAME KheFrameActiveAtRight(KHE_FRAME frame, KHE_RESOURCE r);
extern KHE_FRAME KheFrameInactiveAtRight(KHE_FRAME frame, KHE_RESOURCE r);
*** */

/* ***
extern void KheFrameIteratorInit(KHE_FRAME_ITERATOR fi, KHE_FRAME frame,
  KHE_RESOURCE r, int extra);
extern bool KheFrameIteratorNext(KHE_FRAME_ITERATOR fi, KHE_FRAME *res);
extern bool KheFrameIteratorNextPair(KHE_FRAME_ITERATOR fi, KHE_FRAME *res);
*** */

/* 12.5.1 Satisfying requested task assignments */
extern bool KheSolnAssignRequestedResources(KHE_SOLN soln, KHE_OPTIONS options);

/* 12.5.2 Most-constrained-first assignment */
extern bool KheMostConstrainedFirstAssignResources(KHE_TASKING tasking,
  KHE_OPTIONS options);

/* 12.5.3 Resource packing */
extern bool KheResourcePackAssignResources(KHE_TASKING tasking,
  KHE_OPTIONS options);

/* 12.5.4 Split assignments */
extern bool KheFindSplitResourceAssignments(KHE_TASKING tasking,
  KHE_OPTIONS options);

/* 12.6.1 A solver for resource matching */
extern KHE_RESOURCE_MATCHING_SOLVER KheResourceMatchingSolverMake(
  KHE_SOLN soln, KHE_RESOURCE_GROUP rg, HA_ARENA a);
extern void KheResourceMatchingSolverDelete(KHE_RESOURCE_MATCHING_SOLVER rms);

extern bool KheResourceMatchingSolverSolve(KHE_RESOURCE_MATCHING_SOLVER rms,
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, bool edge_adjust1_off,
  bool edge_adjust2_off, bool edge_adjust3_off, bool edge_adjust4_off,
  bool nocost_off, KHE_OPTIONS options);
extern bool KheResourceMatchingSolverSolveWithLookahead(
  KHE_RESOURCE_MATCHING_SOLVER rms,
  ARRAY_KHE_RESOURCE_MATCHING_DEMAND_SET *rmds_array,
  int first_index, int last_index, bool edge_adjust1_off,
  bool edge_adjust2_off, bool edge_adjust3_off,
  bool edge_adjust4_off, bool nocost_off, KHE_OPTIONS options);

extern KHE_RESOURCE_MATCHING_DEMAND_SET KheResourceMatchingDemandSetMake(
  KHE_RESOURCE_MATCHING_SOLVER rms, bool preserve_existing);
extern void KheResourceMatchingDemandSetAddTime(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_TIME t);
extern void KheResourceMatchingDemandSetAddTimeGroup(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_TIME_GROUP tg);
extern void KheResourceMatchingDemandSetAddFrame(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_FRAME frame);
extern void KheResourceMatchingDemandSetDelete(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds);
extern void KheResourceMatchingDemandSetClear(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds);

/* 12.6.3 Time sweep resource assignment */
extern bool KheTimeSweepAssignResources(KHE_SOLN soln, KHE_RESOURCE_GROUP rg,
  KHE_OPTIONS options);

/* 12.6.4 Resource rematching repair */
extern bool KheResourceRematch(KHE_SOLN soln, KHE_RESOURCE_GROUP rg,
  KHE_OPTIONS options, int variant);

/* 12.8 Resource pair repair */
extern bool KheResourcePairReassign(KHE_SOLN soln, KHE_RESOURCE r1,
  KHE_RESOURCE r2, bool resource_invariant, bool fix_splits);
extern bool KheResourcePairRepair(KHE_TASKING tasking, KHE_OPTIONS options);

extern bool KheResourcePairSimpleReassign(KHE_SOLN soln,
  KHE_RESOURCE r1, KHE_RESOURCE r2, KHE_FRAME frame, int fi, int li,
  bool resource_invariant, int max_assignments);
extern bool KheResourcePairSimpleRepair(KHE_SOLN soln, KHE_OPTIONS options);
extern bool KheResourcePairSimpleBusyRepair(KHE_SOLN soln, KHE_OPTIONS options);

extern bool KheResourceTripleSimpleReassign(KHE_SOLN soln,
  KHE_RESOURCE r1, KHE_RESOURCE r2, KHE_RESOURCE r3, KHE_FRAME frame,
  int fi, int li, bool resource_invariant, int max_assignments);
extern bool KheResourceTripleSimpleRepair(KHE_SOLN soln, KHE_OPTIONS options);

extern bool KheResourcePairSwap(KHE_SOLN soln, KHE_RESOURCE r1,
  KHE_RESOURCE r2, KHE_FRAME frame, int fi, int li, bool resource_invariant);
extern bool KheResourcePairSwapRepair(KHE_SOLN soln, KHE_OPTIONS options);

/* 12.9 Trying unassignments */
extern bool KheSolnTryTaskUnAssignments(KHE_SOLN soln);

/* 12.10 Putting it all together */
extern bool KheTaskingAssignResourcesStage1(KHE_TASKING tasking,
  KHE_OPTIONS options);
extern bool KheTaskingAssignResourcesStage2(KHE_TASKING tasking,
  KHE_OPTIONS options);
extern bool KheTaskingAssignResourcesStage3(KHE_TASKING tasking,
  KHE_OPTIONS options);
/* extern void KheAssignResourcesGlobalCheck(KHE_SOLN soln, char *pos); */


/*****************************************************************************/
/*                                                                           */
/*    Chapter 13.  Ejection Chains                                           */
/*                                                                           */
/*****************************************************************************/

/* 13.2 Ejector construction */
extern KHE_EJECTOR KheEjectorMakeBegin(HA_ARENA a);
extern void KheEjectorMakeEnd(KHE_EJECTOR ej);
extern HA_ARENA KheEjectorArena(KHE_EJECTOR ej);
/* extern void KheEjectorDelete(KHE_EJECTOR ej); */

/* 13.2 Ejector construction - major schedules */
extern int KheEjectorMajorScheduleCount(KHE_EJECTOR ej);
extern KHE_EJECTOR_MAJOR_SCHEDULE KheEjectorMajorSchedule(KHE_EJECTOR ej,
  int i);

extern void KheEjectorMajorScheduleBegin(KHE_EJECTOR ej);
extern void KheEjectorMajorScheduleEnd(KHE_EJECTOR ej);
extern int KheEjectorMajorScheduleMinorScheduleCount(
  KHE_EJECTOR_MAJOR_SCHEDULE ejm);
extern KHE_EJECTOR_MINOR_SCHEDULE KheEjectorMajorScheduleMinorSchedule(
  KHE_EJECTOR_MAJOR_SCHEDULE ejm, int i);

/* 13.2 Ejector construction - minor schedules */
extern void KheEjectorMinorScheduleAdd(KHE_EJECTOR ej,
  int max_depth, bool may_revisit);
extern int KheEjectorMinorScheduleMaxDepth(KHE_EJECTOR_MINOR_SCHEDULE ejms);
extern bool KheEjectorMinorScheduleMayRevisit(KHE_EJECTOR_MINOR_SCHEDULE ejms);

/* 13.2 Ejector construction - other schedules */
extern void KheEjectorAddDefaultSchedules(KHE_EJECTOR ej);
extern void KheEjectorSetSchedulesFromString(KHE_EJECTOR ej,
  char *ejector_schedules_string);

/* 13.2 Ejector construction - augment functions */
extern void KheEjectorAddAugment(KHE_EJECTOR ej, KHE_MONITOR_TAG tag,
  KHE_EJECTOR_AUGMENT_FN augment_fn, int augment_type);
extern void KheEjectorAddGroupAugment(KHE_EJECTOR ej, int sub_tag,
  KHE_EJECTOR_AUGMENT_FN augment_fn, int augment_type);

/* 13.3 Ejector solving */
extern bool KheEjectorSolve(KHE_EJECTOR ej, KHE_GROUP_MONITOR start_gm,
  KHE_GROUP_MONITOR continue_gm, KHE_OPTIONS options);
extern void KheEjectorSolveBegin(KHE_EJECTOR ej, KHE_GROUP_MONITOR start_gm,
  KHE_GROUP_MONITOR continue_gm, KHE_OPTIONS options);
extern bool KheEjectorSolveEnd(KHE_EJECTOR ej);

/* 13.3 Ejector solving - cost limits */
extern void KheEjectorAddMonitorCostLimit(KHE_EJECTOR ej,
  KHE_MONITOR m, KHE_COST cost_limit);
extern void KheEjectorAddMonitorCostLimitReducing(KHE_EJECTOR ej,
  KHE_MONITOR m);
extern int KheEjectorMonitorCostLimitCount(KHE_EJECTOR ej);
extern void KheEjectorMonitorCostLimit(KHE_EJECTOR ej, int i,
  KHE_MONITOR *m, KHE_COST *cost_limit, bool *reducing);

/* 13.3 Ejector solving - queries */
extern KHE_GROUP_MONITOR KheEjectorStartGroupMonitor(KHE_EJECTOR ej);
extern KHE_GROUP_MONITOR KheEjectorContinueGroupMonitor(KHE_EJECTOR ej);
extern KHE_OPTIONS KheEjectorOptions(KHE_EJECTOR ej);
extern KHE_FRAME KheEjectorFrame(KHE_EJECTOR ej);
extern KHE_EVENT_TIMETABLE_MONITOR KheEjectorEventTimetableMonitor(
  KHE_EJECTOR ej);
extern KHE_SOLN KheEjectorSoln(KHE_EJECTOR ej);
extern KHE_COST KheEjectorTargetCost(KHE_EJECTOR ej);
extern KHE_EJECTOR_MAJOR_SCHEDULE KheEjectorCurrMajorSchedule(KHE_EJECTOR ej);
extern KHE_EJECTOR_MINOR_SCHEDULE KheEjectorCurrMinorSchedule(KHE_EJECTOR ej);
extern bool KheEjectorCurrMayRevisit(KHE_EJECTOR ej);
extern int KheEjectorCurrDepth(KHE_EJECTOR ej);
extern int KheEjectorCurrAugmentCount(KHE_EJECTOR ej);
extern bool KheEjectorCurrDebug(KHE_EJECTOR ej);
extern int KheEjectorCurrDebugIndent(KHE_EJECTOR ej);

/* 13.3 Ejector solving - repairs */
extern void KheEjectorRepairBegin(KHE_EJECTOR ej);
extern bool KheEjectorRepairEndLong(KHE_EJECTOR ej, int repair_type,
  bool success, int max_sub_chains, bool save_and_sort,
  void (*on_success_fn)(void *on_success_val), void *on_success_val);
extern bool KheEjectorRepairEnd(KHE_EJECTOR ej, int repair_type, bool success);

/* 13.6.3 Statistics describing a single solve */
extern int KheEjectorImprovementCount(KHE_EJECTOR ej);
extern int KheEjectorImprovementNumberOfRepairs(KHE_EJECTOR ej, int i);
extern float KheEjectorImprovementTime(KHE_EJECTOR ej, int i);
extern KHE_COST KheEjectorImprovementCost(KHE_EJECTOR ej, int i);
extern int KheEjectorImprovementDefects(KHE_EJECTOR ej, int i);

extern KHE_COST KheEjectorInitCost(KHE_EJECTOR ej);
extern int KheEjectorInitDefects(KHE_EJECTOR ej);

/* 13.6.4 Statistics describing multiple solves */
extern int KheEjectorImprovementRepairHistoMax(KHE_EJECTOR ej);
extern int KheEjectorImprovementRepairHistoFrequency(KHE_EJECTOR ej,
  int repair_count);
extern int KheEjectorImprovementRepairHistoTotal(KHE_EJECTOR ej);
extern float KheEjectorImprovementRepairHistoAverage(KHE_EJECTOR ej);

extern int KheEjectorImprovementAugmentHistoMax(KHE_EJECTOR ej);
extern int KheEjectorImprovementAugmentHistoFrequency(KHE_EJECTOR ej,
  int augment_count);
extern int KheEjectorImprovementAugmentHistoTotal(KHE_EJECTOR ej);
extern float KheEjectorImprovementAugmentHistoAverage(KHE_EJECTOR ej);

extern int KheEjectorTotalRepairs(KHE_EJECTOR ej, int augment_type);
extern int KheEjectorTotalRepairsByType(KHE_EJECTOR ej, int augment_type,
  int repair_type);
extern int KheEjectorSuccessfulRepairs(KHE_EJECTOR ej, int augment_type);
extern int KheEjectorSuccessfulRepairsByType(KHE_EJECTOR ej, int augment_type,
  int repair_type);

extern void KheEjectorAddAugmentType(KHE_EJECTOR ej, int augment_type,
  char *augment_label);
extern int KheEjectorAugmentTypeCount(KHE_EJECTOR ej);
extern int KheEjectorAugmentType(KHE_EJECTOR ej, int i);
extern char *KheEjectorAugmentTypeLabel(KHE_EJECTOR ej, int augment_type);

extern void KheEjectorAddRepairType(KHE_EJECTOR ej, int repair_type,
  char *repair_label);
extern int KheEjectorRepairTypeCount(KHE_EJECTOR ej);
extern int KheEjectorRepairType(KHE_EJECTOR ej, int i);
extern char *KheEjectorRepairTypeLabel(KHE_EJECTOR ej, int repair_type);

/* 13.7 Ejection chain time and resource repair functions */
extern bool KheEjectionChainNodeRepairTimes(KHE_NODE parent_node,
  KHE_OPTIONS options);
extern bool KheEjectionChainLayerRepairTimes(KHE_LAYER layer,
  KHE_OPTIONS options);
extern bool KheEjectionChainRepairResources(KHE_TASKING tasking,
  KHE_OPTIONS options);
extern bool KheEjectionChainRepairInitialResourceAssignment(
  KHE_GROUP_MONITOR limit_resources_gm, KHE_OPTIONS options);

extern KHE_EJECTOR KheEjectionChainEjectorMake(KHE_OPTIONS options,
  HA_ARENA a);
extern KHE_EJECTOR KheEjectionChainEjectorOption(KHE_OPTIONS options,
  char *key);

extern void KheEjectionChainPrepareMonitors(KHE_SOLN soln, KHE_OPTIONS options);
extern void KheEjectionChainUnPrepareMonitors(KHE_SOLN soln);

extern KHE_GROUP_MONITOR KheNodeTimeRepairStartGroupMonitorMake(KHE_NODE node);
extern KHE_GROUP_MONITOR KheLayerTimeRepairStartGroupMonitorMake(
  KHE_LAYER layer);
extern KHE_GROUP_MONITOR KheLayerTimeRepairLongStartGroupMonitorMake(
  KHE_LAYER layer);
extern KHE_GROUP_MONITOR KheTaskingStartGroupMonitorMake(KHE_TASKING tasking);
extern KHE_GROUP_MONITOR KheGroupEventMonitors(KHE_SOLN soln,
  KHE_MONITOR_TAG tag, KHE_SUBTAG_STANDARD_TYPE sub_tag);
extern KHE_GROUP_MONITOR KheAllDemandGroupMonitorMake(KHE_SOLN soln);

#endif
