
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_combined.c                                          */
/*  DESCRIPTION:  Combined resource solvers                                  */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>

#define DEBUG1 0	/* entry and exit of each main call */
#define DEBUG2 0	/* stages within each main call     */
#define DEBUG3 0	/* monitor debug within repair      */
#define DEBUG4 0	/* KheSolnCheckTasks                */
#define DEBUG5 0	/* repair steps                     */
#define DEBUG6 0
/* ***
#define DEBUG7 1
#define DEBUG7_ID "DemandConstraint:18A/1Wed:E"
*** */
#define DEBUG8 0	/* clash checking */
#define DEBUG9 0
#define DEBUG10 0


/*****************************************************************************/
/*                                                                           */
/*  Submodule "checks"                                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheSolnCheckTasks(KHE_SOLN soln, char *label)                       */
/*                                                                           */
/*  Check soln for inconsistent tasks; if found, print an error message      */
/*  including label.                                                         */
/*                                                                           */
/*****************************************************************************/

static void KheSolnCheckTasks(KHE_SOLN soln, char *label)
{
  int i;  KHE_TASK task;  KHE_RESOURCE ass_r, pre_r;  KHE_EVENT e;
  char *role;  KHE_EVENT_RESOURCE er;  KHE_INSTANCE ins;
  for( i = 0;  i < KheSolnTaskCount(soln);  i++ )
  {
    task = KheSolnTask(soln, i);

    /* only interested in tasks derived from event resources */
    er = KheTaskEventResource(task);
    if( er == NULL )
      continue;

    /* get assigned and preassigned resources and check that they have Ids */
    role = KheEventResourceRole(er);
    e = KheEventResourceEvent(er);
    ass_r = KheTaskAsstResource(task);
    pre_r = KheEventResourcePreassignedResource(er);
    ins = KheEventInstance(e);
    HnAssert(ass_r == NULL || KheResourceId(ass_r) != NULL,
      "KheSolnCheckTasks %s: in instance %s, resource without Id assigned to "
      "event %s (role %s)", label, KheInstanceId(ins), KheEventId(e),
      role == NULL ? "(none)" : role);
    HnAssert(pre_r == NULL || KheResourceId(pre_r) != NULL,
      "KheSolnCheckTasks %s: in instance %s, resource without Id preassigned to"
      " event %s (role %s)", label, KheInstanceId(ins), KheEventId(e),
      role == NULL ? "(none)" : role);

    if( pre_r != NULL )
    {
      /* if preassigned, check that the assigned resource is equal to it */
      HnAssert(ass_r != NULL,
	"KheSolnCheckTasks %s: in event %s of instance %s, event resource with "
	"preassigned resource %s has task with missing resource assignment",
	label, KheEventId(e), KheInstanceId(ins), KheResourceId(pre_r));
      HnAssert(ass_r == pre_r,
	"KheSolnCheckTasks %s: in event %s of instance %s, event resource with "
	"preassigned resource %s has task with inconsistent assignment %s",
	label, KheEventId(e), KheInstanceId(ins), KheResourceId(pre_r),
	KheResourceId(ass_r));
    }
    else
    {
      /* if unpreassigned, must have a role */
      HnAssert(role != NULL, "KheSolnCheckTasks %s: in event %s of instance %s, "
	"unpreassigned event resource has no role", label, KheEventId(e),
	KheInstanceId(ins));
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheNoClashesCheck(KHE_SOLN soln, KHE_RESOURCE r)                    */
/*                                                                           */
/*  Make sure that r has no clashes.                                         */
/*                                                                           */
/*****************************************************************************/

static void KheNoClashesCheck(KHE_SOLN soln, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  if( r != NULL )
  {
    rtm = KheResourceTimetableMonitor(soln, r);
    HnAssert(KheResourceTimetableMonitorClashingTimeCount(rtm) == 0,
      "KheNoClashesCheck failed on resource %s", KheResourceId(r));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSolnCheckForClashes(KHE_SOLN soln)                               */
/*                                                                           */
/*  Check for clashes.                                                       */
/*                                                                           */
/*****************************************************************************/

static void KheSolnCheckForClashes(KHE_SOLN soln)
{
  int i;  KHE_RESOURCE r;
  KHE_INSTANCE ins = KheSolnInstance(soln);
  for( i = 0;  i < KheInstanceResourceCount(ins);  i++ )
  {
    r = KheInstanceResource(ins, i);
    KheNoClashesCheck(soln, r);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "trying unassignments"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheSolnTryTaskUnAssignments(KHE_SOLN soln)                          */
/*                                                                           */
/*  Try unassigning each task of soln, to see if that improves the cost.     */
/*  Return true if any unassignments were kept.                              */
/*                                                                           */
/*****************************************************************************/

bool KheSolnTryTaskUnAssignments(KHE_SOLN soln)
{
  KHE_TASK task, target_task;  int i, j;  KHE_COST init_cost;  bool ok, res;
  KHE_TRACE trace;  KHE_MONITOR m;  KHE_RESOURCE r;
  if( DEBUG9 )
    fprintf(stderr, "[ KheSolnTryTaskUnAssignments(soln)\n");
  res = false;
  for( i = 0;  i < KheSolnTaskCount(soln);  i++ )
  {
    task = KheSolnTask(soln, i);
    target_task = KheTaskAsst(task);
    init_cost = KheSolnCost(soln);
    if( target_task != NULL && !KheTaskIsPreassigned(task, NULL) )
    {
      if( DEBUG10 )
      {
	r = KheTaskAsstResource(task);
	trace = KheTraceMake((KHE_GROUP_MONITOR) soln);
	KheTraceBegin(trace);
	ok = KheTaskUnAssign(task);
	KheTraceEnd(trace);
	fprintf(stderr, "  unassign %s from ",
	  r == NULL ? "?" : KheResourceId(r));
	KheTaskDebug(task, 1, -1, stderr);
	fprintf(stderr, ":  %.5f -> %.5f\n",
	  KheCostShow(init_cost), KheCostShow(KheSolnCost(soln)));
	for( j = 0;  j < KheTraceMonitorCount(trace);  j++ )
	{
	  m = KheTraceMonitor(trace, j);
	  fprintf(stderr, "      from %.5f to ",
	    KheCostShow(KheTraceMonitorInitCost(trace, j)));
	  KheMonitorDebug(m, 2, 0, stderr);
	}
	KheTraceDelete(trace);
      }
      else
        ok = KheTaskUnAssign(task);
      if( ok )
      {
	if( KheSolnCost(soln) < init_cost )
	{
	  res = true;
	  if( DEBUG9 )
	  {
	    fprintf(stderr, "  %.5f -> %.5f for unassigning task ",
	      KheCostShow(init_cost), KheCostShow(KheSolnCost(soln)));
	    KheTaskDebug(task, 1, 0, stderr);
	  }
	}
	else
	  KheTaskAssign(task, target_task);
      }
    }
  }
  if( DEBUG9 )
    fprintf(stderr, "] KheSolnTryTaskUnAssignments returning %s\n",
      res ? "true" : "false");
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "putting it all together"                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  char *KheTaskingLabel(KHE_TASKING tasking)                               */
/*                                                                           */
/*  Return a string description of the scope of tasking.                     */
/*                                                                           */
/*****************************************************************************/

static char *KheTaskingLabel(KHE_TASKING tasking)
{
  return KheTaskingResourceType(tasking) == NULL ? "All Resources" :
    KheResourceTypeId(KheTaskingResourceType(tasking)) == NULL ? "?" :
    KheResourceTypeId(KheTaskingResourceType(tasking));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDebugStage(KHE_TASKING tasking, char *stage, int verbosity)      */
/*                                                                           */
/*  Print a debug message showing the stage, and the solution cost.          */
/*                                                                           */
/*****************************************************************************/

static void KheDebugStage(KHE_TASKING tasking, char *stage, int verbosity)
{
  fprintf(stderr, "  %s:\n", stage);
  KheSolnDebug(KheTaskingSoln(tasking), verbosity, 2, stderr);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceDebugMonitor(KHE_OPTIONS options, char *intro,           */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug the nominated monitor.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheResourceDebugMonitor(KHE_OPTIONS options, char *intro,
  int indent, FILE *fp)
{
  KHE_MONITOR m;
  m = (KHE_MONITOR) KheOptionsGetObject(options, "gs_debug_monitor", NULL);
  if( m != NULL )
  {
    fprintf(fp, "%*s%s: ", indent, "", intro);
    KheMonitorDebug(m, 1, 0, fp);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void DebugStart(KHE_SOLN soln, KHE_OPTIONS options, char *fn_name)       */
/*                                                                           */
/*  Brief report of what is starting, for debugging.                         */
/*                                                                           */
/*****************************************************************************/

/* *** replaced by KheSolveDebug
static void DebugStart(KHE_SOLN soln, KHE_OPTIONS options, char *fn_name)
{
  KHE_INSTANCE ins;  KHE_TIMER timer;  char buff[20];
  if( DEBUG5 )
  {
    ins = KheSolnInstance(soln);
    fprintf(stderr, "%s div %d", KheInstanceId(ins), KheSolnDiversifier(soln));
    if( KheOptionsContainsTimer(options, "global", &timer) )
      fprintf(stderr, " after %s",
	KheTimeShow(KheTimerElapsedTime(timer), buff));
    fprintf(stderr, " calling %s\n", fn_name);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheAssignResourcesGlobalCheck(KHE_SOLN soln, char *pos)             */
/*                                                                           */
/*  Global check of soln.                                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheAssignResourcesGlobalCheck(KHE_SOLN soln, char *pos)
{
  KHE_MONITOR m;
  if( DEBUG7 )
  {
    if( !KheSolnRetrieveMonitor(soln, DEBUG7_ID, &m) )
      HnAbort("DebugCheck (khe_sr_combined.c) internal error 1");
    HnAssert(KheMonitorTag(m) == KHE_LIMIT_RESOURCES_MONITOR_TAG,
      "DebugCheck (khe_sr_combined.c) internal error 2");
    KheLimitResourcesMonitorCheck((KHE_LIMIT_RESOURCES_MONITOR) m, pos);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskingRepairResources(KHE_TASKING tasking, KHE_OPTIONS options, */
/*    float time_limit_multiplier, int variant)                              */
/*                                                                           */
/*  Repair the assignments of tasking using all the tricks available.        */
/*  (Not forgetting that each repair algorithm can be turned off by          */
/*  an option.)  Return true if any improvement.                             */
/*                                                                           */
/*  The time limit (if present) is multiplied by  time_limit_multiplier.     */
/*                                                                           */
/*  Parameter variant is passed to resource rematch to make it vary          */
/*  its behaviour on different calls on this function.                       */
/*                                                                           */
/*****************************************************************************/

static bool KheTaskingRepairResources(KHE_TASKING tasking, KHE_OPTIONS options,
  float time_limit_multiplier, int variant)
{
  KHE_COST init_cost;  KHE_SOLN soln;  KHE_INSTANCE ins;  KHE_RESOURCE_TYPE rt;
  float repair_time_limit;  KHE_TIMER repair_timer, rematch_timer;
  soln = KheTaskingSoln(tasking);
  if( DEBUG2 )
    KheDebugStage(tasking, "at start of KheTaskingRepairResources", 2);
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  repair_time_limit = time_limit_multiplier * KheTimeFromString(
    KheOptionsGet(options, "rs_repair_time_limit", "-"));
  if( repair_time_limit >= 0.0 )
  {
    repair_timer = KheOptionsAddTimer(options, "repair", repair_time_limit);
    rematch_timer = KheOptionsAddTimer(options, "rematch",
      repair_time_limit / 2.0);
  }
  else
    repair_timer = rematch_timer = NULL;
  init_cost = KheSolnCost(soln);
  ins = KheSolnInstance(soln);
  rt = KheTaskingResourceType(tasking);
  if( DEBUG5 )
    KheSolveDebug(soln, options, "KheResourceRematch");
  /* DebugStart(soln, options, "KheResourceRematch"); */
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  if( !KheOptionsGetBool(options, "rs_repair_rematch_off", false) )
    KheResourceRematch(soln, KheResourceTypeFullResourceGroup(rt), options,
      variant);
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  if( rematch_timer != NULL )
    KheOptionsDeleteTimer(options, rematch_timer);
  if( DEBUG4 )
    KheSolnCheckTasks(KheTaskingSoln(tasking), "after KheResourceRematch");
  if( DEBUG5 )
    KheSolveDebug(soln, options, "KheEjectionChainRepairResources");
  /* DebugStart(soln, options, "KheEjectionChainRepairResources"); */
  if( !KheOptionsGetBool(options, "rs_repair_ejection_off", false) )
    KheEjectionChainRepairResources(tasking, options);
  if( DEBUG4 )
    KheSolnCheckTasks(KheTaskingSoln(tasking),
      "after KheEjectionChainRepairResources");
  if( KheInstanceModel(ins) == KHE_MODEL_EMPLOYEE_SCHEDULE )
  {
    if( DEBUG5 )
      KheSolveDebug(soln, options, "KheResourcePairSimpleRepair");
    /* DebugStart(soln, options, "KheResourcePairSimpleRepair"); */
    KheResourcePairSimpleRepair(soln, options);
    if( DEBUG4 )
      KheSolnCheckTasks(KheTaskingSoln(tasking),
	"after KheResourcePairSimpleRepair");
    if( DEBUG5 )
      KheSolveDebug(soln, options, "KheResourcePairSimpleBusyRepair");
    /* DebugStart(soln, options, "KheResourcePairSimpleBusyRepair"); */
    KheResourcePairSimpleBusyRepair(soln, options);
    if( DEBUG4 )
      KheSolnCheckTasks(KheTaskingSoln(tasking),
	"after KheResourcePairSimpleBusyRepair");
    /* DebugStart(soln, options, "KheResourceTripleSimpleRepair"); */
    KheResourceTripleSimpleRepair(soln, options);
    if( DEBUG4 )
      KheSolnCheckTasks(KheTaskingSoln(tasking),
	"after KheResourceTripleSimpleRepair");
    KheResourcePairSwapRepair(soln, options);
  }
  if( repair_timer != NULL )
    KheOptionsDeleteTimer(options, repair_timer);
  if( DEBUG6 && KheSolnCost(soln) < init_cost )
    fprintf(stderr, "KheTaskingRepairResources improvement (%.5f -> %.5f)\n",
      KheCostShow(init_cost), KheCostShow(KheSolnCost(soln)));
  return KheSolnCost(soln) < init_cost;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDoTimeSweep(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,                 */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Carry out a time sweep assignment, along with a few other odd things.    */
/*                                                                           */
/*****************************************************************************/

static void KheDoTimeSweep(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,
  KHE_OPTIONS options)
{
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  if( KheResourceTypeLimitResourcesCount(rt) > 0 )
  {
    KheOptionsSetBool(options, "rs_time_sweep_matching_off", true);
    KheOptionsSetBool(options, "rs_rematch_matching_off", true);
  }
  if( DEBUG5 )
    KheSolveDebug(soln, options, "KheSolnAssignRequestedResources");
  /* DebugStart(soln, options, "KheSolnAssignRequestedResources"); */
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  KheSolnAssignRequestedResources(soln, options);
  if( DEBUG4 )
    KheSolnCheckTasks(soln, "after Stage1 requested");
  if( DEBUG5 )
    KheSolveDebug(soln, options, "KheTimeSweepAssignResources");
  /* DebugStart(soln, options, "KheTimeSweepAssignResources"); */
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  KheTimeSweepAssignResources(soln, KheResourceTypeFullResourceGroup(rt),
    options);
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  if( DEBUG4 )
    KheSolnCheckTasks(soln, "after Stage1 time sweep");
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskingAssignResourcesStage1(KHE_TASKING tasking,                */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Assign resources to the leader tasks of tasking (Stage 1):  assign       */
/*  most tasks preserving the resource assignment invariant.                 */
/*                                                                           */
/*****************************************************************************/

bool KheTaskingAssignResourcesStage1(KHE_TASKING tasking, KHE_OPTIONS options)
{
  KHE_RESOURCE_TYPE rt;  KHE_INSTANCE ins;  char *resource_assign;
  KHE_SOLN soln;  KHE_TASK_SET group_by_rc_ts, group_by_resource_ts;
  if( DEBUG1 )
    fprintf(stderr, "[ KheTaskingAssignResourcesStage1(%s)\n",
      KheTaskingLabel(tasking));
  soln = KheTaskingSoln(tasking);
  if( DEBUG4 )
    KheSolnCheckTasks(soln, "at start of Stage1");
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);

  /* maintain the resource assignment invariant */
  KheOptionsSetBool(options, "rs_invariant", true);

  /* grouping by frame */
  if( DEBUG5 )
    KheSolveDebug(soln, options, "KheTaskingGroupByResourceConstraints");
  /* DebugStart(soln, options, "KheTaskingGroupByResourceConstraints"); */
  group_by_rc_ts = KheTaskSetMake(soln);
  KheGroupByResourceConstraints(soln, KheTaskingResourceType(tasking),
    options, group_by_rc_ts);
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);

  /* sort out the resource_assign options */
  ins = KheSolnInstance(soln);
  rt = KheTaskingResourceType(tasking);
  resource_assign = KheOptionsGet(options, "rs_constructor", "auto");
  if( strcmp(resource_assign, "auto") == 0 )
  {
    if( KheInstanceModel(ins) == KHE_MODEL_EMPLOYEE_SCHEDULE )
      resource_assign = "time_sweep";
    else if( rt != NULL && KheResourceTypeAvoidSplitAssignmentsCount(rt) == 0 )
      resource_assign = "most_constrained";
    else
      resource_assign = "resource_packing";
  }

  /* construction */
  if( DEBUG5 )
    KheSolveDebug(soln, options, "%s", resource_assign);
  /* DebugStart(soln, options, resource_assign); */
  if( strcmp(resource_assign, "none") == 0 )
  {
    /* do nothing */
  }
  else if( strcmp(resource_assign, "most_constrained") == 0 )
    KheMostConstrainedFirstAssignResources(tasking, options);
  else if( strcmp(resource_assign, "resource_packing") == 0 )
    KheResourcePackAssignResources(tasking, options);
  else if( strcmp(resource_assign, "time_sweep") == 0 )
    KheDoTimeSweep(soln, rt, options);
  else
    HnAbort("KheTaskingAssignResourcesStage1: "
      "invalid resource_assign option \"%s\"", resource_assign);
  if( DEBUG8 )
    KheSolnCheckForClashes(soln);
  if( DEBUG4 )
    KheSolnCheckTasks(KheTaskingSoln(tasking), "after Stage1 construction");

  /* ungroup the grouping by frame */
  /* *** prefer to do it after first repair
  KheTaskSetUnGroup(group_by_rc_ts);
  KheTaskSetDelete(group_by_rc_ts);
  *** */

  /* repair */
  if( !KheOptionsGetBool(options, "rs_repair_off", false) &&
      !KheOptionsGetBool(options, "rs_repair1_off", false) )
  {
    /* repair without grouping by resource */
    if( DEBUG3 )
      KheResourceDebugMonitor(options, "before Stage1 repair", 2, stderr);
    KheTaskingRepairResources(tasking, options, 1.0, 0);

    /* optional repair with grouping by resource */
    if( KheOptionsGetBool(options, "rs_group_by_resource", false) )
    {
      group_by_resource_ts = KheTaskSetMake(soln);
      if( KheTaskingGroupByResource(tasking, options, group_by_resource_ts) )
	KheTaskingRepairResources(tasking, options, 0.5, 1);
      KheTaskSetUnGroup(group_by_resource_ts);
      KheTaskSetDelete(group_by_resource_ts);
    }
  }

  /* ungroup the grouping by frame */
  KheTaskSetUnGroup(group_by_rc_ts);
  KheTaskSetDelete(group_by_rc_ts);

  if( DEBUG1 )
  {
    KheDebugStage(tasking, "at end of Stage1", 2);
    fprintf(stderr, "] KheTaskingAssignResourcesStage1 returning\n");
  }
  if( DEBUG4 )
    KheSolnCheckTasks(KheTaskingSoln(tasking), "at end of Stage1");
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskingAssignResourcesStage2(KHE_TASKING tasking,                */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Assign resources to the leader tasks of tasking (Stage 2):  find         */
/*  split assignments, still preserving the resource assignment invariant.   */
/*                                                                           */
/*****************************************************************************/

bool KheTaskingAssignResourcesStage2(KHE_TASKING tasking, KHE_OPTIONS options)
{
  KHE_RESOURCE_TYPE rt;

  rt = KheTaskingResourceType(tasking);
  if( KheResourceTypeAvoidSplitAssignmentsCount(rt) > 0 &&
      !KheOptionsGetBool(options, "rs_repair_off", false) &&
      !KheOptionsGetBool(options, "rs_repair2_off", false) )
  {
    if( DEBUG1 )
      fprintf(stderr, "[ KheTaskingAssignResourcesStage2(%s)\n",
	KheTaskingLabel(tasking));
    KheFindSplitResourceAssignments(tasking, options);
    KheTaskingAllowSplitAssignments(tasking, false);
    if( DEBUG3 )
      KheResourceDebugMonitor(options, "monitor before Stage2 repair",2,stderr);
    KheTaskingRepairResources(tasking, options, 0.5, 0);
    if( DEBUG1 )
    {
      KheDebugStage(tasking, "at end of Stage2", 2);
      fprintf(stderr, "] KheTaskingAssignResourcesStage2 returning\n");
    }
  }
  if( DEBUG4 )
    KheSolnCheckTasks(KheTaskingSoln(tasking), "at end of Stage2");
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskingAssignResourcesStage3(KHE_TASKING tasking,                */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Assign resources to the leader tasks of tasking (Stage 2):  last-ditch   */
/*  assignments, without the resource assignment invariant.                  */
/*                                                                           */
/*****************************************************************************/

bool KheTaskingAssignResourcesStage3(KHE_TASKING tasking, KHE_OPTIONS options)
{
  KHE_SOLN soln;
  if( !KheOptionsGetBool(options, "rs_repair_off", false) &&
      !KheOptionsGetBool(options, "rs_repair3_off", false) )
  {
    if( DEBUG1 )
      fprintf(stderr, "[ KheTaskingAssignResourcesStage3(%s)\n",
	KheTaskingLabel(tasking));
    soln = KheTaskingSoln(tasking);
    if( DEBUG8 )
      KheSolnCheckForClashes(soln);
    KheOptionsSetBool(options, "rs_invariant", false);
    KheTaskingEnlargeDomains(tasking, true);
    if( DEBUG3 )
      KheResourceDebugMonitor(options, "before Stage3 repair", 2, stderr);
    /* KheOptionsSetBool(options, "es_fresh_visits", true); */
    KheTaskingRepairResources(tasking, options, 0.5, 1);
    /* ***
    if( DEBUG4 )
      KheSolnCheckTasks(soln, "second repair in Stage3");
    KheTaskingRepairResources(tasking, options, 0.5, 1);
    *** */
    if( DEBUG4 )
      KheSolnCheckTasks(soln, "at end of Stage3");
    if( DEBUG1 )
    {
      KheDebugStage(tasking, "at end of Stage3", 2);
      fprintf(stderr, "] KheTaskingAssignResourcesStage3 returning\n");
    }
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskingAssignResources(KHE_TASKING tasking, KHE_OPTIONS options) */
/*                                                                           */
/*  Assign resources to the leader tasks of tasking.                         */
/*                                                                           */
/*****************************************************************************/

/* *** not used; was not in khe.h
bool KheTaskingAssignResources(KHE_TASKING tasking, KHE_OPTIONS options)
{
  KheTaskingAssignResourcesStage1(tasking, options);
  KheTaskingAssignResourcesStage2(tasking, options);
  KheTaskingAssignResourcesStage3(tasking, options);
  return true;
}
*** */
