
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU Generar Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_resource_matching.c                                 */
/*  DESCRIPTION:  Resource matching                                          */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "khe_mmatch.h"
#include <limits.h>

#define bool_show(a) ((a) ? "true" : "false")

#define DEBUG1 0	/* trace best asst                                   */
#define DEBUG3 0	/* individual solves                                 */
#define DEBUG4 0	/* make best asst                                    */
#define DEBUG5 0	/* make best asst                                    */
#define DEBUG6 0	/* equivalence                                       */
#define DEBUG7 0	/* graph construction                                */
#define DEBUG8 0	/* adding tasks                                      */
#define DEBUG9 0	/* tracing during graph construction                 */
#define DEBUG10 0	/* initial/final cost problem                        */


/*****************************************************************************/
/*                                                                           */
/*  KHE_SUPPLY_NODE                                                          */
/*                                                                           */
/*  One supply node sn of the graph.                                         */
/*                                                                           */
/*  If sn->resource != NULL, then sn represents assigning sn->resource to    */
/*  one or all of the tasks of some (any) demand node.  The multiplicity     */
/*  of sn is 1, because assigning sn->resource to more than one node would   */
/*  cause a clash.                                                           */
/*                                                                           */
/*  If sn->resource == NULL, then sn represents not assigning a resource     */
/*  to one or all of the tasks of some (any) demand node.  The multiplicity  */
/*  of sn is effectively inifinite, because assigning nothing can be done    */
/*  many times.  However, the MMatch module does not accept INT_MAX, so we   */
/*  actually use the total multiplicity of the demand nodes.                 */
/*                                                                           */
/*  Fields c1, c2, c3, and c4 are only used by the heuristic solver, only    */
/*  temporarily, while choosing the best assignments for one demand node.    */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_supply_node_rec {
  KHE_RESOURCE_MATCHING_SOLVER	solver;		/* enclosing solver          */
  KHE_RESOURCE			resource;	/* the resource (optional)   */
  int				max_busy;	/* max resource busy times   */
  int				curr_busy;	/* reset for each match  */
  /* int			curr_at_max; */	/* reset for each match  */
  bool				curr_was_used;	/* reset for each match  */
  int				curr_run_len;	/* reset for each match  */
  KHE_MMATCH_NODE		mmatch_node;	/* supply node in mmatch     */
  int				c1;		/* temp for heuristic solve  */
  int				c2;		/* temp for heuristic solve  */
  int				c3;		/* temp for heuristic solve  */
  int				c4;		/* temp for heuristic solve  */
} *KHE_SUPPLY_NODE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DEMAND_NODE                                                          */
/*                                                                           */
/*  One demand node dn of the graph, representing some included tasks.       */
/*                                                                           */
/*  If dn->grouped, its tasks were all initially  assigned to the same       */
/*  resource, so they are to be assigned to the same resource in the         */
/*  result.  So the node's multiplicity is 1:  just one matching edge        */
/*  should touch it.                                                         */
/*                                                                           */
/*  If !dn->grouped, the tasks were all initially unassigned, and they are   */
/*  equivalent.  So they may be assigned to different resources, and the     */
/*  multiplicity of the node is equal to the number of tasks.                */
/*                                                                           */
/*  The duration is the duration that will be added to a resource that       */
/*  matches with this node.  If the node is grouped it is the total          */
/*  duration of the tasks.  If the node is not grouped it is the duration    */
/*  of any one of the (equivalent) tasks.                                    */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_TASK) ARRAY_KHE_TASK;

typedef struct khe_demand_node_rec {
  KHE_RESOURCE_MATCHING_SOLVER	solver;		/* enclosing solver          */
  bool				grouped;	/* true if grouped           */
  ARRAY_KHE_TASK		tasks;		/* the tasks                 */
  KHE_TIME			time;		/* any time tasks running    */
  int				duration;	/* task duration             */
  KHE_MMATCH_NODE		mmatch_node;	/* demand node in mmatch     */
} *KHE_DEMAND_NODE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_MATCHING_DEMAND_SET                                         */
/*                                                                           */
/*  A set of demand nodes to be solved, with associated attributes.          */
/*                                                                           */
/*  The demand_nodes array holds all the demand nodes.                       */
/*                                                                           */
/*  The resource_demand_nodes array has one entry for each resource in       */
/*  KheResourceGroupResourceType(resource_group).  The entry at index        */
/*  KheResourceResourceTypeIndex(r) is the demand node holding the           */
/*  included tasks which are either preassigned r or initially assigned      */
/*  r.  If there are no such tasks, either because r is not in the resource  */
/*  group or it is but is not assigned or preassigned to any included tasks, */
/*  there is no such node and the entry is NULL.                             */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_DEMAND_NODE) ARRAY_KHE_DEMAND_NODE;
typedef HA_ARRAY(KHE_TIME) ARRAY_KHE_TIME;

struct khe_resource_matching_demand_set_rec {
  KHE_RESOURCE_MATCHING_SOLVER	solver;			/* solver            */
  ARRAY_KHE_TIME		times;			/* times             */
  bool				ready_for_solving;	/* ready for solving */
  bool				preserve_existing;	/* preserve existing */
  ARRAY_KHE_TASK		tasks;			/* the tasks         */
  ARRAY_KHE_DEMAND_NODE		demand_nodes;		/* the demand nodes  */
  ARRAY_KHE_DEMAND_NODE		resource_demand_nodes;  /* indexed by resrce */
  char				*debug_string;		/* demand set's name */
};

typedef HA_ARRAY(KHE_RESOURCE_MATCHING_DEMAND_SET)
  ARRAY_KHE_RESOURCE_MATCHING_DEMAND_SET;


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_MATCHING_SOLVER                                             */
/*                                                                           */
/*  A resource matching solver.                                              */
/*                                                                           */
/*  The supply_nodes array holds all the supply nodes, one for each of the   */
/*  resources of the resource group, plus one representing non-assignment.   */
/*  These are fixed throughout the lifetime of the solver, although the      */
/*  heuristic solver does permute their order.                               */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_SUPPLY_NODE) ARRAY_KHE_SUPPLY_NODE;

struct khe_resource_matching_solver_rec {

  /* fields with a well-defined value for the lifetime of the solver */
  HA_ARENA			arena;			/* arena             */
  KHE_SOLN			soln;			/* the solution      */
  KHE_RESOURCE_GROUP		resource_group;		/* resources affected*/
  bool				max_busy_set;		/* max_busy set      */
  KHE_MMATCH			mmatch;			/* match object      */
  ARRAY_KHE_SUPPLY_NODE		supply_nodes;		/* supply nodes      */
  ARRAY_KHE_DEMAND_NODE		free_demand_nodes;	/* free demand nodes */
  ARRAY_KHE_RESOURCE_MATCHING_DEMAND_SET free_demand_sets;

  /* fields with a well-defined value during a single solve */
  KHE_RESOURCE_MATCHING_DEMAND_SET demand_set;		/* demand set        */
  KHE_OPTIONS			options;		/* options           */
  KHE_FRAME			common_frame;		/* common frame      */
  bool				resource_invariant;	/* preserve invt     */
  bool				matching_off;		/* don't use mmatch  */
  bool				edge_adjust_off;	/* don't adjust edges*/
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "supply nodes"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_SUPPLY_NODE KheSupplyNodeMake(KHE_RESOURCE_MATCHING_SOLVER rms,      */
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Make a new supply node and add it to rms->supply_nodes.  If r != NULL,   */
/*  the node represents assignment to r.  Otherwise the node represents      */
/*  the non-assignment of any demand node.                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_SUPPLY_NODE KheSupplyNodeMake(KHE_RESOURCE_MATCHING_SOLVER rms,
  KHE_RESOURCE r)
{
  KHE_SUPPLY_NODE res;
  HaMake(res, rms->arena);
  res->solver = rms;
  res->resource = r;
  res->max_busy = 0;  /* will be revised later if required */
  res->curr_busy = 0;    /* reset for each match  */
  /* res->curr_at_max = 0; */  /* reset for each match  */
  res->curr_was_used = false;	/* reset for each match  */
  res->curr_run_len = 0;	/* reset for each match  */
  res->mmatch_node = NULL;
  res->c1 = 0;
  res->c2 = 0;
  res->c3 = 0;
  res->c4 = 0;
  HaArrayAddLast(rms->supply_nodes, res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheSupplyNodeMultiplicity(KHE_SUPPLY_NODE sn)                        */
/*                                                                           */
/*  Return the multiplicity of sn.  If it is a resource node, this is        */
/*  just 1.  Otherwise it is the total multiplicity of all demand nodes.     */
/*                                                                           */
/*****************************************************************************/
static int KheDemandNodeMultiplicity(KHE_DEMAND_NODE dn);

static int KheSupplyNodeMultiplicity(KHE_SUPPLY_NODE sn)
{
  int res, i;  KHE_DEMAND_NODE dn;  KHE_RESOURCE_MATCHING_DEMAND_SET rmds;
  if( sn->resource != NULL )
    return 1;
  else
  {
    rmds = sn->solver->demand_set;
    res = 0;
    HaArrayForEach(rmds->demand_nodes, dn, i)
      res += KheDemandNodeMultiplicity(dn);
    return res;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSupplyNodeAddMMatchNode(KHE_SUPPLY_NODE sn, KHE_MMATCH m)        */
/*                                                                           */
/*  Add a mmatch node corresponding to sn to m.                              */
/*                                                                           */
/*****************************************************************************/

static void KheSupplyNodeAddMMatchNode(KHE_SUPPLY_NODE sn, KHE_MMATCH m)
{
  sn->mmatch_node = KheMMatchSupplyNodeMake(m, KheSupplyNodeMultiplicity(sn),
    (void *) sn);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheResourceBusyTimes(KHE_SOLN soln, KHE_RESOURCE r)                  */
/*                                                                           */
/*  Return the number of times that r is currently busy in soln.             */
/*                                                                           */
/*****************************************************************************/

static int KheResourceBusyTimes(KHE_SOLN soln, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  rtm = KheResourceTimetableMonitor(soln, r);
  return KheResourceTimetableMonitorBusyTimes(rtm);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheSupplyNodeTypedIncreasingCostCmp(KHE_SUPPLY_NODE sn1,             */
/*    KHE_SUPPLY_NODE sn2)                                                   */
/*                                                                           */
/*  Comparison function for sorting supply nodes by increasing (c1,c2,c3,c4) */
/*  - only used by the heuristic solver.                                     */
/*                                                                           */
/*****************************************************************************/

static int KheSupplyNodeTypedIncreasingCostCmp(KHE_SUPPLY_NODE sn1,
  KHE_SUPPLY_NODE sn2)
{
  if( sn1->c1 != sn2->c1 )
    return sn1->c1 - sn2->c1;
  if( sn1->c2 != sn2->c2 )
    return sn1->c2 - sn2->c2;
  if( sn1->c3 != sn2->c3 )
    return sn1->c3 - sn2->c3;
  return sn1->c4 - sn2->c4;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheSupplyNodeLT(KHE_SUPPLY_NODE sn1, KHE_SUPPLY_NODE sn2)           */
/*                                                                           */
/*  Return true if the cost in sn1 is less than or equal to the cost in sn2. */
/*                                                                           */
/*****************************************************************************/

static bool KheSupplyNodeLT(KHE_SUPPLY_NODE sn1, KHE_SUPPLY_NODE sn2)
{
  if( sn1->c1 != sn2->c1 )
    return sn1->c1 < sn2->c1;
  if( sn1->c2 != sn2->c2 )
    return sn1->c2 < sn2->c2;
  if( sn1->c3 != sn2->c3 )
    return sn1->c3 < sn2->c3;
  return sn1->c4 < sn2->c4;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheSupplyNodeIncreasingCostCmp(const void *t1, const void *t2)       */
/*                                                                           */
/*  Comparison function for sorting supply nodes by increasing (c1,c2,c3,c4) */
/*  - only used by the heuristic solver.                                     */
/*                                                                           */
/*****************************************************************************/

int KheSupplyNodeIncreasingCostCmp(const void *t1, const void *t2)
{
  KHE_SUPPLY_NODE sn1 = * (KHE_SUPPLY_NODE *) t1;
  KHE_SUPPLY_NODE sn2 = * (KHE_SUPPLY_NODE *) t2;
  return KheSupplyNodeTypedIncreasingCostCmp(sn1, sn2);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSupplyNodeDebug(KHE_SUPPLY_NODE sn, int verbosity,               */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug function for printing a supply node.   It may be NULL, denoting    */
/*  non-assignment.                                                          */
/*                                                                           */
/*****************************************************************************/

static void KheSupplyNodeDebug(KHE_SUPPLY_NODE sn, int verbosity,
  int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( sn->resource == NULL )
    fprintf(fp, "noassign");
  else
  {
    fprintf(fp, "%s", KheResourceId(sn->resource));
    if( !sn->solver->edge_adjust_off )
      fprintf(fp, ":mb%dcb%d", sn->max_busy, sn->curr_busy);
      /* ***
      fprintf(fp, ":mb%dcb%dcam%dcrl%d", sn->max_busy, sn->curr_busy,
	sn->curr_at_max, sn->curr_run_len);
      *** */
  }
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSupplyNodeUnTypedDebug(void *value, int verbosity,               */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug function for printing a supply node.                               */
/*                                                                           */
/*****************************************************************************/

static void KheSupplyNodeUnTypedDebug(void *value, int verbosity,
  int indent, FILE *fp)
{
  KheSupplyNodeDebug((KHE_SUPPLY_NODE) value, verbosity, indent, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "demand nodes"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DEMAND_NODE KheDemandNodeMake(KHE_RESOURCE_MATCHING_DEMAND_SET rmds, */
/*    bool grouped)                                                          */
/*                                                                           */
/*  Make a new demand node, representing an initially empty set of included  */
/*  tasks, and add it to rmds->demand_nodes.                                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DEMAND_NODE KheDemandNodeMake(KHE_RESOURCE_MATCHING_DEMAND_SET rmds,
  bool grouped)
{
  KHE_DEMAND_NODE res;  KHE_RESOURCE_MATCHING_SOLVER rms;  HA_ARENA a;
  rms = rmds->solver;
  if( HaArrayCount(rms->free_demand_nodes) > 0 )
  {
    res = HaArrayLastAndDelete(rms->free_demand_nodes);
    HaArrayClear(res->tasks);
  }
  else
  {
    a = rmds->solver->arena;
    HaMake(res, a);
    HaArrayInit(res->tasks, a);
  }
  res->solver = rms;
  res->grouped = grouped;
  res->time = NULL;
  res->duration = 0;
  res->mmatch_node = NULL;
  HaArrayAddLast(rmds->demand_nodes, res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDemandNodeAddTask(KHE_DEMAND_NODE dn, KHE_TASK task)             */
/*                                                                           */
/*  Add task to dn.                                                          */
/*                                                                           */
/*****************************************************************************/

static void KheDemandNodeAddTask(KHE_DEMAND_NODE dn, KHE_TASK task)
{
  HaArrayAddLast(dn->tasks, task);
  if( dn->grouped || HaArrayCount(dn->tasks) == 1 )
    dn->duration += KheTaskTotalDuration(task);
  if( dn->time == NULL && KheTaskMeet(task) != NULL )
    dn->time = KheMeetAsstTime(KheTaskMeet(task));
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandNodeMultiplicity(KHE_DEMAND_NODE dn)                        */
/*                                                                           */
/*  Return the multiplicity of dn.                                           */
/*                                                                           */
/*****************************************************************************/

static int KheDemandNodeMultiplicity(KHE_DEMAND_NODE dn)
{
  return (dn->grouped ? 1 : HaArrayCount(dn->tasks));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDemandNodeAddMMatchNode(KHE_DEMAND_NODE dn, KHE_MMATCH m)        */
/*                                                                           */
/*  Add a mmatch node corresponding to dn to m.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDemandNodeAddMMatchNode(KHE_DEMAND_NODE dn, KHE_MMATCH m)
{
  dn->mmatch_node = KheMMatchDemandNodeMake(m, KheDemandNodeMultiplicity(dn),
    (void *) dn);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDemandNodeAssignAllTasks(KHE_DEMAND_NODE dn, KHE_RESOURCE r)     */
/*                                                                           */
/*  Try to assign r to all of the tasks of dn, optionally preserving the     */
/*  resource assignment invariant.  Return true if all the tasks are         */
/*  assigned r afterwards.                                                   */
/*                                                                           */
/*****************************************************************************/

static bool KheDemandNodeAssignAllTasks(KHE_DEMAND_NODE dn, KHE_RESOURCE r)
{
  KHE_RESOURCE_MATCHING_SOLVER rms;  int dc, i;
  KHE_TASK task;  bool all_assigned;
  rms = dn->solver;
  dc = KheSolnMatchingDefectCount(rms->soln);
  all_assigned = true;
  HaArrayForEach(dn->tasks, task, i)
    if( !KheEjectingTaskMoveFrame(task, r, false, rms->common_frame) )
      all_assigned = false;
  return all_assigned &&
    (!rms->resource_invariant || KheSolnMatchingDefectCount(rms->soln) <= dc);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDemandNodeAssignOneTask(KHE_DEMAND_NODE dn, KHE_TASK task,       */
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Try to assign r to task (a task from dn), optionally preserving the      */
/*  resource assignment invariant.  Return true if task is assigned r        */
/*  afterwards.                                                              */
/*                                                                           */
/*****************************************************************************/

static bool KheDemandNodeAssignOneTask(KHE_DEMAND_NODE dn, KHE_TASK task,
  KHE_RESOURCE r)
{
  KHE_RESOURCE_MATCHING_SOLVER rms;  int dc;  bool all_assigned;
  rms = dn->solver;
  dc = KheSolnMatchingDefectCount(rms->soln);
  all_assigned = KheEjectingTaskMoveFrame(task, r, false, rms->common_frame);
  return all_assigned &&
    (!rms->resource_invariant || KheSolnMatchingDefectCount(rms->soln) <= dc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDemandNodeUnAssignTasks(KHE_DEMAND_NODE dn)                      */
/*                                                                           */
/*  Unassign the assigned tasks of dn.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheDemandNodeUnAssignTasks(KHE_DEMAND_NODE dn)
{
  int i;  KHE_TASK task;
  HaArrayForEach(dn->tasks, task, i)
    if( KheTaskAsstResource(task) != NULL )
      KheTaskUnAssign(task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDemandNodeDebug(KHE_DEMAND_NODE dn, int verbosity,               */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug function for printing a demand node.                               */
/*                                                                           */
/*****************************************************************************/

static void KheDemandNodeDebug(KHE_DEMAND_NODE dn, int verbosity,
  int indent, FILE *fp)
{
  KHE_TASK task;  int i;
  if( indent >= 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[%s", dn->grouped ? "grouped " : "");
  HaArrayForEach(dn->tasks, task, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    KheTaskDebug(task, 2, -1, fp);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDemandNodeUnTypedDebug(void *value, int verbosity,               */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug function for printing a demand node.                               */
/*                                                                           */
/*****************************************************************************/

static void KheDemandNodeUnTypedDebug(void *value, int verbosity,
  int indent, FILE *fp)
{
  KheDemandNodeDebug((KHE_DEMAND_NODE) value, verbosity, indent, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "demand sets"                                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheTimeIncreasingIndexCmp(const void *p1, const void *p2)            */
/*                                                                           */
/*  Comparison function for sorting an array of times into increasing        */
/*  index order.                                                             */
/*                                                                           */
/*****************************************************************************/

static int KheTimeIncreasingIndexCmp(const void *p1, const void *p2)
{
  KHE_TIME time1 = * (KHE_TIME *) p1;
  KHE_TIME time2 = * (KHE_TIME *) p2;
  return KheTimeIndex(time1) - KheTimeIndex(time2);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheTaskIncreasingSolnIndexCmp(const void *p1, const void *p2)        */
/*                                                                           */
/*  Comparison functions for sorting an array of tasks by increasing         */
/*  KheTaskSolnIndex().                                                      */
/*                                                                           */
/*****************************************************************************/

static int KheTaskIncreasingSolnIndexCmp(const void *p1, const void *p2)
{
  KHE_TASK task1 = * (KHE_TASK *) p1;
  KHE_TASK task2 = * (KHE_TASK *) p2;
  return KheTaskSolnIndex(task1) - KheTaskSolnIndex(task2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskEquivalent(KHE_TASK task1, KHE_TASK task2)                   */
/*                                                                           */
/*  Return true if task1 and task2 are equivalent, that is, if assigning     */
/*  task1 to a resource must have the same effect on its timetable, and      */
/*  on solution cost, as assigning task2.                                    */
/*                                                                           */
/*  We already know that task1 and task2 are unassigned, so we don't         */
/*  check that again.                                                        */
/*                                                                           */
/*****************************************************************************/

static bool KheTaskEquivalent(KHE_TASK task1, KHE_TASK task2)
{
  KHE_EVENT_RESOURCE er1, er2;  KHE_MEET meet1, meet2;
  KHE_TASK child_task1, child_task2;  int i;

  /* must be derived from equivalent event resources */
  er1 = KheTaskEventResource(task1);
  er2 = KheTaskEventResource(task2);
  if( er1 == NULL || er2 == NULL || !KheEventResourceEquivalent(er1, er2) )
    return false;

  /* must have the same duration and the same starting time */
  meet1 = KheTaskMeet(task1);
  meet2 = KheTaskMeet(task2);
  if( KheMeetDuration(meet1) != KheMeetDuration(meet2) )
    return false;
  if( KheMeetAsstTime(meet1) != KheMeetAsstTime(meet2) )
    return false;

  /* must have the same domain */
  if( !KheResourceGroupEqual(KheTaskDomain(task1), KheTaskDomain(task2)) )
    return false;

  /* assigned tasks must be equivalent */
  /* we should really sort them first, but we don't */
  if( KheTaskAssignedToCount(task1) != KheTaskAssignedToCount(task2) )
    return false;
  for( i = 0;  i < KheTaskAssignedToCount(task1);  i++ )
  {
    child_task1 = KheTaskAssignedTo(task1, i);
    child_task2 = KheTaskAssignedTo(task2, i);
    if( !KheTaskEquivalent(child_task1, child_task2) )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingDemandSetPrepareForSolving(                      */
/*    KHE_RESOURCE_MATCHING_DEMAND_SET rmds)                                 */
/*                                                                           */
/*  Make rmds ready for solving.  This is done only once, the first time     */
/*  rdms is solved, unless rmds is cleared or more times are added.          */
/*                                                                           */
/*****************************************************************************/

static void KheResourceMatchingDemandSetPrepareForSolving(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds)
{
  KHE_TASK task, prev_ungrouped_task;
  KHE_RESOURCE_MATCHING_SOLVER rms;  KHE_DEMAND_NODE dn, prev_ungrouped_dn;
  KHE_EVENT_TIMETABLE_MONITOR etm;
  KHE_TIME t;  int i, j, k, ri;  KHE_RESOURCE r;
  KHE_MEET meet;  KHE_RESOURCE_TYPE rt;  bool group_is_all;

  /* boilerplate */
  if( DEBUG8 )
    fprintf(stderr, "[ KheResourceMatchingDemandSetPrepareForSolving:\n");
  HnAssert(HaArrayCount(rmds->times) > 0,
    "KheResourceMatchingSolverSolve: no times");
  rms = rmds->solver;
  etm = (KHE_EVENT_TIMETABLE_MONITOR) KheOptionsGetObject(rms->options,
    "gs_event_timetable_monitor", NULL);
  rt = KheResourceGroupResourceType(rms->resource_group);
  group_is_all = (KheResourceTypeResourceCount(rt) ==
    KheResourceGroupResourceCount(rms->resource_group));

  /* get the tasks */
  HaArraySortUnique(rmds->times, &KheTimeIncreasingIndexCmp);
  HaArrayForEach(rmds->times, t, i)
  {
    if( DEBUG8 )
      fprintf(stderr, "  examining time %s\n", KheTimeId(t));
    for( j = 0;  j < KheEventTimetableMonitorTimeMeetCount(etm, t);  j++ )
    {
      meet = KheEventTimetableMonitorTimeMeet(etm, t, j);
      if( DEBUG8 )
      {
	fprintf(stderr, "  examining meet ");
	KheMeetDebug(meet, 1, -1, stderr);
	fprintf(stderr, "  at time %s\n", KheTimeId(t));
      }
      for( k = 0;  k < KheMeetTaskCount(meet);  k++ )
      {
	task = KheMeetTask(meet, k);
	if( KheTaskResourceType(task) == rt )
	{
	  if( DEBUG8 )
	  {
	    fprintf(stderr, "    examining task ");
	    KheTaskDebug(task, 1, 0, stderr);
	  }
	  task = KheTaskProperRoot(task);
	  if( DEBUG8 )
	  {
	    fprintf(stderr, "      proper root is ");
	    KheTaskDebug(task, 1, -1, stderr);
	    if( KheTaskIsPreassigned(task, &r) )
	      fprintf(stderr, ", preassigned %s", KheResourceId(r));
	    if( KheTaskAssignIsFixed(task) )
	      fprintf(stderr, ", fixed");
	    if( KheTaskAsst(task) != NULL )
	    {
	      fprintf(stderr, ", assigned to ");
	      KheTaskDebug(KheTaskAsst(task), 1, -1, stderr);
	    }
	    fprintf(stderr, "\n");
	    fprintf(stderr, "      condition1: %s && %s && %s\n",
              !KheTaskIsPreassigned(task, &r) ? "true" : "false",
              !KheTaskAssignIsFixed(task) ? "true" : "false",
              (!rmds->preserve_existing || KheTaskAsst(task) == NULL) ?
	        "true" : "false");
	  }
	  if( !KheTaskIsPreassigned(task, &r) && !KheTaskAssignIsFixed(task)
	      && (!rmds->preserve_existing || KheTaskAsst(task) == NULL) )
	  {
	    r = KheTaskAsstResource(task);
	    if( DEBUG8 )
	      fprintf(stderr, "      condition2: %s\n", r==NULL || group_is_all
		|| KheResourceGroupContains(rms->resource_group, r) ?
		"true" : "false");
	    if( r == NULL || group_is_all ||
		KheResourceGroupContains(rms->resource_group, r) )
	    {
	      if( DEBUG8 )
		fprintf(stderr, "      condition3: %s\n",
		  KheTaskNonAssignmentHasCost(task) ? "true" : "false");
	      if( KheTaskNonAssignmentHasCost(task) )
	      {
		if( DEBUG8 )
		{
		  fprintf(stderr, "      adding task ");
		  KheTaskDebug(task, 1, 0, stderr);
		}
		HaArrayAddLast(rmds->tasks, task);
	      }
	    }
	  }
	}
      }
    }
  }
  HaArraySortUnique(rmds->tasks, &KheTaskIncreasingSolnIndexCmp);
  if( DEBUG8 )
    fprintf(stderr, "  %d tasks after uniquifying\n",HaArrayCount(rmds->tasks));

  /* build the demand nodes, adding the tasks to them */
  prev_ungrouped_task = NULL;  prev_ungrouped_dn = NULL;
  HaArrayForEach(rmds->tasks, task, i)
  {
    /* set r to the resource task is assigned to, or NULL */
    r = KheTaskAsstResource(task);
    if( r != NULL )
    {
      /* use r's grouped demand node; it may need to be created */
      ri = KheResourceResourceTypeIndex(r);
      HaArrayFill(rmds->resource_demand_nodes, ri + 1, NULL);
      dn = HaArray(rmds->resource_demand_nodes, ri);
      if( dn == NULL )
      {
	dn = KheDemandNodeMake(rmds, true);
	HaArrayPut(rmds->resource_demand_nodes, ri, dn);
      }
      KheDemandNodeAddTask(dn, task);
    }
    else
    {
      /* use an existing or new ungrouped demand node */
      if( prev_ungrouped_task == NULL ||
	  !KheTaskEquivalent(prev_ungrouped_task, task) )
	prev_ungrouped_dn = KheDemandNodeMake(rmds, false);
      KheDemandNodeAddTask(prev_ungrouped_dn, task);
      prev_ungrouped_task = task;
    }
  }
  if( DEBUG8 )
    fprintf(stderr,
      "] KheResourceMatchingDemandSetPrepareForSolving (%d demands)\n",
      HaArrayCount(rmds->demand_nodes));
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_MATCHING_DEMAND_SET KheResourceMatchingDemandSetMake(       */
/*    KHE_RESOURCE_MATCHING_SOLVER rms, bool preserve_existing)              */
/*                                                                           */
/*  Make and return a new demand set for rms, initially with no times.       */
/*                                                                           */
/*****************************************************************************/

KHE_RESOURCE_MATCHING_DEMAND_SET KheResourceMatchingDemandSetMake(
  KHE_RESOURCE_MATCHING_SOLVER rms, bool preserve_existing)
{
  KHE_RESOURCE_MATCHING_DEMAND_SET res;
  if( HaArrayCount(rms->free_demand_sets) > 0 )
  {
    res = HaArrayLastAndDelete(rms->free_demand_sets);
    HaArrayClear(res->times);
    HaArrayClear(res->demand_nodes);
    HaArrayClear(res->resource_demand_nodes);
    HaArrayClear(res->tasks);
  }
  else
  {
    HaMake(res, rms->arena);
    HaArrayInit(res->times, rms->arena);
    HaArrayInit(res->demand_nodes, rms->arena);
    HaArrayInit(res->resource_demand_nodes, rms->arena);
    HaArrayInit(res->tasks, rms->arena);
  }
  res->solver = rms;
  res->ready_for_solving = false;
  res->preserve_existing = preserve_existing;
  res->debug_string = NULL;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingDemandSetAddTime(                                */
/*    KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_TIME t)                     */
/*                                                                           */
/*  Add time t to rmds.                                                      */
/*                                                                           */
/*****************************************************************************/

void KheResourceMatchingDemandSetAddTime(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_TIME t)
{
  if( rmds->ready_for_solving )
    KheResourceMatchingDemandSetClear(rmds);
  HaArrayAddLast(rmds->times, t);
  if( rmds->debug_string == NULL )
    rmds->debug_string = KheTimeId(t);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingDemandSetAddTimeGroup(                           */
/*    KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_TIME_GROUP tg)              */
/*                                                                           */
/*  Add the times of tg to rmds.                                             */
/*                                                                           */
/*****************************************************************************/

void KheResourceMatchingDemandSetAddTimeGroup(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_TIME_GROUP tg)
{
  KHE_TIME t;  int i;
  if( rmds->ready_for_solving )
    KheResourceMatchingDemandSetClear(rmds);
  for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
  {
    t = KheTimeGroupTime(tg, i);
    HaArrayAddLast(rmds->times, t);
  }
  if( rmds->debug_string == NULL )
    rmds->debug_string = KheTimeGroupId(tg);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingDemandSetAddFrame(                               */
/*    KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_FRAME frame)                */
/*                                                                           */
/*  Add the times of the time groups of frame to rmds.                       */
/*                                                                           */
/*****************************************************************************/

void KheResourceMatchingDemandSetAddFrame(
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, KHE_FRAME frame)
{
  int i;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  if( rmds->ready_for_solving )
    KheResourceMatchingDemandSetClear(rmds);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i, &po);
    KheResourceMatchingDemandSetAddTimeGroup(rmds, tg);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingDemandSetDelete(                                 */
/*    KHE_RESOURCE_MATCHING_DEMAND_SET rmds)                                 */
/*                                                                           */
/*  Delete rmds.                                                             */
/*                                                                           */
/*****************************************************************************/

void KheResourceMatchingDemandSetDelete(KHE_RESOURCE_MATCHING_DEMAND_SET rmds)
{
  int i;
  HaArrayAppend(rmds->solver->free_demand_nodes, rmds->demand_nodes, i);
  HaArrayClear(rmds->demand_nodes);
  HaArrayAddLast(rmds->solver->free_demand_sets, rmds);
  /* ***
  MArrayFree(rmds->times);
  MArrayFree(rmds->tasks);
  MArrayFree(rmds->demand_nodes);
  MArrayFree(rmds->resource_demand_nodes);
  MFree(rmds);
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingDemandSetClear(                                  */
/*    KHE_RESOURCE_MATCHING_DEMAND_SET rmds)                                 */
/*                                                                           */
/*  Clear rmds back to its state when it was first made.                     */
/*                                                                           */
/*****************************************************************************/

void KheResourceMatchingDemandSetClear(KHE_RESOURCE_MATCHING_DEMAND_SET rmds)
{
  int i;
  rmds->ready_for_solving = false;
  HaArrayClear(rmds->times);
  HaArrayClear(rmds->tasks);
  HaArrayAppend(rmds->solver->free_demand_nodes, rmds->demand_nodes, i);
  HaArrayClear(rmds->demand_nodes);
  HaArrayClear(rmds->resource_demand_nodes);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "solver construction and solving"                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingSolverBackDebug(void *value, int verbosity,      */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug function for printing the back value of a time sweep mmatch.       */
/*                                                                           */
/*****************************************************************************/

static void KheResourceMatchingSolverBackDebug(void *value, int verbosity,
  int indent, FILE *fp)
{
  KHE_RESOURCE_MATCHING_SOLVER rms;
  rms = (KHE_RESOURCE_MATCHING_SOLVER) value;
  fprintf(fp, "%s", rms->demand_set->debug_string == NULL ? "?" :
    rms->demand_set->debug_string);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingCostDebug(void *back, int64_t cost, FILE *fp)    */
/*                                                                           */
/*  Debug function for displaying cost.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheResourceMatchingCostDebug(int c1, int c2, int c3, FILE *fp)
{
  if( c2 > 99999 )
    fprintf(fp, "%d,%d", c1, c2);
  else
    fprintf(fp, "%.5f", KheCostShow(KheCost(c1, c2)));
  if( c3 != 0 )
    fprintf(fp, "|%d", c3);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingEdgeDebug(KHE_DEMAND_NODE dn,                    */
/*    KHE_SUPPLY_NODE sn, int c1, int c2, int c3, int indent, FILE *fp)      */
/*                                                                           */
/*  Debug print of an edge from dn to sn with cost (c1, c2, c3).             */
/*                                                                           */
/*****************************************************************************/

static void KheResourceMatchingEdgeDebug(KHE_DEMAND_NODE dn,
  KHE_SUPPLY_NODE sn, int mult, int c1, int c2, int c3, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  KheDemandNodeDebug(dn, 1, -1, fp);
  if( mult != INT_MAX )
    fprintf(fp, " %d--", mult);
  else
    fprintf(fp, " --");
  KheResourceMatchingCostDebug(c1, c2, c3, fp);
  fprintf(fp, "-> ");
  KheSupplyNodeDebug(sn, 1, -1, fp);
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  int KheExtra(KHE_TASK task, KHE_RESOURCE r)                              */
/*                                                                           */
/*  A bit extra, based on the diversifier and the task and resource          */
/*  indexes, just to shake things up a bit.                                  */
/*                                                                           */
/*****************************************************************************/

/* *** this just seems to make things slightly worse
static int KheExtra(KHE_TASK task, KHE_RESOURCE r, int range)
{
  int diversifier;
  diversifier = KheSolnDiversifier(KheTaskSoln(task));
  return (diversifier * 5 + KheTaskSolnIndex(task) * 7 +
    KheResourceInstanceIndex(r) * 13) % range;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheResourceAtMaxCount(KHE_SOLN soln, KHE_RESOURCE r)                 */
/*                                                                           */
/*  Return the total at max count of r in soln, but only if there are        */
/*  not too many monitors to hunt through.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheResourceAtMaxCount(KHE_SOLN soln, KHE_RESOURCE r)
{
  int i, res, count;  KHE_MONITOR m;
  res = 0;
  count = KheSolnResourceMonitorCount(soln, r);
  if( count < 300 )
    for( i = 0;  i < count;  i++ )
    {
      m = KheSolnResourceMonitor(soln, r, i);
      if( KheMonitorTag(m) == KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG )
	res += KheClusterBusyTimesMonitorAtMaxLimitCount(
	  (KHE_CLUSTER_BUSY_TIMES_MONITOR) m);
      else if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
	res += KheLimitActiveIntervalsMonitorAtMaxLimitCount(
	  (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m);
    }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheResourceAtMaxCount(KHE_RESOURCE r, KHE_SOLN soln, KHE_TIME t)     */
/*                                                                           */
/*  Return the at_max value of r at t, or 0 if t is NULL.                    */
/*                                                                           */
/*****************************************************************************/

static int KheResourceAtMaxCount(KHE_RESOURCE r, KHE_SOLN soln, KHE_TIME t)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  if( t == NULL )
    return 0;
  rtm = KheResourceTimetableMonitor(soln, r);
  return KheResourceTimetableMonitorAtMaxLimitCount(rtm, t);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheEdge(KHE_RESOURCE_MATCHING_SOLVER rms, KHE_DEMAND_NODE dn,       */
/*    KHE_SUPPLY_NODE sn, int *c1, int *c2, int *c3)                         */
/*                                                                           */
/*  If there needs to be an edge between demand node dn and supply node sn,  */
/*  return true with (*c1, *c2, *c3) set to its cost, else return false.     */
/*                                                                           */
/*****************************************************************************/
#define max(a, b) ((a) > (b) ? (a) : (b))

static bool KheEdge(KHE_RESOURCE_MATCHING_SOLVER rms, KHE_DEMAND_NODE dn,
  KHE_SUPPLY_NODE sn, int *c1, int *c2, int *c3)
{
  KHE_COST cost;
  int part_count, busy_times, at_max_before,  at_max_after;
  if( sn->resource == NULL )
  {
    /* leave dn unassigned, whatever it is */
    cost = KheSolnCost(rms->soln);
    *c1 = KheHardCost(cost);
    *c2 = KheSoftCost(cost);
    *c3 = 0;
    return true;
  }
  else
  {
    at_max_before = KheResourceAtMaxCount(sn->resource, rms->soln, dn->time);
    if( dn->grouped ? KheDemandNodeAssignAllTasks(dn, sn->resource) :
      KheDemandNodeAssignOneTask(dn, HaArrayFirst(dn->tasks), sn->resource) )
    {
      /* ***
      at_max_before = rms->edge_adjust_off ? 0 :
	KheResourceAtMaxCount(rms->soln, sn->resource);
      *** */

      /* we've assigned one or all of the tasks of dn to sn->resource */
      cost = KheSolnCost(rms->soln);
      *c1 = KheHardCost(cost);
      *c2 = KheSoftCost(cost);
      if( rms->edge_adjust_off )
	*c3 = 0;
      else
      {
	part_count = KheFrameTimeGroupCount(rms->common_frame);
	busy_times = sn->curr_busy + dn->duration;
	at_max_after = KheResourceAtMaxCount(sn->resource, rms->soln, dn->time);
	*c3 = (part_count - sn->max_busy + busy_times) * 2 +
	  max(0, at_max_after - at_max_before) + sn->curr_run_len;
	/* ***
	*c3 = (part_count - sn->max_busy + busy_times) * 2 +
	  max(0, at_max_after - sn->curr_at_max) +
	  sn->curr_run_len;
	*** */
      }
      return true;
    }
    else
      return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheResourceMatchingEdgeFn(KHE_DEMAND_NODE dn,                       */
/*    KHE_SUPPLY_NODE sn, int64_t *cost)                                     */
/*                                                                           */
/*  If there should be an edge between dn and sn in the bipartite graph,     */
/*  return true and set (*c1, *c2, *c3) to its cost.  Otherwise return       */
/*  false and leave (*c1, *c2, *c3) unchanged.                               */
/*                                                                           */
/*****************************************************************************/

static bool KheResourceMatchingEdgeFn(KHE_DEMAND_NODE dn,
  KHE_SUPPLY_NODE sn, int *c1, int *c2, int *c3)
{
  bool res;  int mult;  KHE_MARK mark;  KHE_RESOURCE_MATCHING_SOLVER rms;

  /* boilerplate */
  HnAssert(c1 != NULL, "KheResourceMatchingEdgeFn internal error 1");
  HnAssert(sn != NULL, "KheResourceMatchingEdgeFn internal error 2");
  rms = dn->solver;

  mark = KheMarkBegin(rms->soln);
  if( DEBUG9 )
  {
    KHE_TRACE trace;  KHE_MONITOR m;  int i;
    trace = KheTraceMake((KHE_GROUP_MONITOR) rms->soln);
    KheTraceBegin(trace);
    res = KheEdge(rms, dn, sn, c1, c2, c3);
    KheTraceEnd(trace);
    fprintf(stderr, "    edge ");
    if( res )
    {
      mult = (sn->resource != NULL ? 1 : KheDemandNodeMultiplicity(dn));
      KheResourceMatchingEdgeDebug(dn, sn, mult, *c1, *c2, *c3, -1, stderr);
      fprintf(stderr, ":\n    [\n");
      for( i = 0;  i < KheTraceMonitorCount(trace);  i++ )
      {
	m = KheTraceMonitor(trace, i);
	fprintf(stderr, "      from %.5f to ",
	  KheCostShow(KheTraceMonitorInitCost(trace, i)));
	KheMonitorDebug(m, 2, 0, stderr);
      }
      KheTraceDelete(trace);
      fprintf(stderr, "    ]\n");
    }
    else
    {
      KheDemandNodeDebug(dn, 1, -1, stderr);
      fprintf(stderr, " -> ");
      KheSupplyNodeDebug(sn, 1, -1, stderr);
      fprintf(stderr, " not present\n");
    }
  }
  else
    res = KheEdge(rms, dn, sn, c1, c2, c3);
  KheMarkEnd(mark, true);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_MATCHING_SOLVER KheResourceMatchingSolverMake(KHE_SOLN soln,*/
/*    KHE_RESOURCE_GROUP rg, HA_ARENA a)                                     */
/*                                                                           */
/*  Make a resource matching solver with these attributes.                   */
/*                                                                           */
/*****************************************************************************/

KHE_RESOURCE_MATCHING_SOLVER KheResourceMatchingSolverMake(KHE_SOLN soln,
  KHE_RESOURCE_GROUP rg, HA_ARENA a)
{
  KHE_RESOURCE_MATCHING_SOLVER res;  int i, count, extra;  KHE_RESOURCE r;

  /* fields with a well-defined value for the lifetime of the solver */
  HaMake(res, a);
  res->arena = a;
  res->soln = soln;
  res->resource_group = rg;
  res->max_busy_set = false;
  res->mmatch = KheMMatchMake((void *) res, &KheResourceMatchingSolverBackDebug,
    &KheDemandNodeUnTypedDebug, &KheSupplyNodeUnTypedDebug,
    &KheResourceMatchingCostDebug, a);
  HaArrayInit(res->supply_nodes, a);
  HaArrayInit(res->free_demand_nodes, a);
  HaArrayInit(res->free_demand_sets, a);

  /* fields with a well-defined value during a single solve */
  res->demand_set = NULL;
  res->options = NULL;
  res->common_frame = KheFrameMakeNull();
  res->resource_invariant = false;
  res->matching_off = false;
  res->edge_adjust_off = false;

  /* initialize supply nodes (the same for all solves) */
  count = KheResourceGroupResourceCount(rg);
  extra = 5 * KheSolnDiversifier(soln);
  for( i = 0;  i < KheResourceGroupResourceCount(rg);  i++ )
  {
    r = KheResourceGroupResource(rg, (i + extra) % count);
    KheSupplyNodeMake(res, r);
  }
  KheSupplyNodeMake(res, NULL);  /* denotes non-assignment of any demand node */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingSolverDelete(KHE_RESOURCE_MATCHING_SOLVER rms)   */
/*                                                                           */
/*  Delete rms.                                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheResourceMatchingSolverDelete(KHE_RESOURCE_MATCHING_SOLVER rms)
{
  KheSolnArenaEnd(rms->soln, rms->arena);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandNodeIncreasingDomainSizeCmp(const void *t1, const void *t2) */
/*                                                                           */
/*  Comparison function for sorting an array of demand nodes into increasing */
/*  order of the size of the domain of their first task.                     */
/*                                                                           */
/*****************************************************************************/

static int KheDemandNodeIncreasingDomainSizeCmp(const void *t1, const void *t2)
{
  KHE_DEMAND_NODE dn1 = * (KHE_DEMAND_NODE *) t1;
  KHE_DEMAND_NODE dn2 = * (KHE_DEMAND_NODE *) t2;
  KHE_TASK task1 = HaArrayFirst(dn1->tasks);
  KHE_TASK task2 = HaArrayFirst(dn2->tasks);
  int size1 = KheResourceGroupResourceCount(KheTaskDomain(task1));
  int size2 = KheResourceGroupResourceCount(KheTaskDomain(task2));
  return size1 - size2;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskMonitoredByAssignOrLimitResourceMonitors(KHE_TASK task)      */
/*                                                                           */
/*  Return true if task is monitored by at least one unsatisfied assign      */
/*  resource or limit resources monitor.                                     */
/*                                                                           */
/*****************************************************************************/

/* *** moved into the platform under the name KheTaskNonAssignmentHasCost
   ** NB this version has at least two bugs:  testing MonitorCost instead
   ** of ConstraintWeight, and not testing descendant tasks as well as task
static bool KheTaskMonitoredByAssignOrLimitResourceMonitors(KHE_TASK task)
{
  KHE_EVENT_RESOURCE er;  KHE_SOLN soln;  int i;  KHE_MONITOR m;
  KHE_MONITOR_TAG tag;

  ** check the task itself **
  soln = KheTaskSoln(task);
  er = KheTaskEventResource(task);
  if( er != NULL )
    for( i = 0;  i < KheSolnEventResourceMonitorCount(soln, er);  i++ )
    {
      m = KheSolnEventResourceMonitor(soln, er, i);
      tag = KheMonitorTag(m);
      if( tag == KHE_ASSIGN_RESOURCE_MONITOR_TAG ||
	  tag == KHE_LIMIT_RESOURCES_MONITOR_TAG )
      {
	if( KheMonitorCost(m) > 0 )
	  return true;
      }
    }

  ** check the tasks assigned to task **
  return false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDemandNodeNonAssignmentHasCost(KHE_DEMAND_NODE dn)               */
/*                                                                           */
/*  Return true if not assigning the tasks of dn would have a cost.          */
/*                                                                           */
/*****************************************************************************/

static bool KheDemandNodeNonAssignmentHasCost(KHE_DEMAND_NODE dn)
{
  KHE_TASK task;  int i;
  HaArrayForEach(dn->tasks, task, i)
    if( KheTaskNonAssignmentHasCost(task) )
      return true;
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_SUPPLY_NODE KheFindBestSupplyNode(KHE_RESOURCE_MATCHING_SOLVER rms,  */
/*    KHE_DEMAND_NODE dn)                                                    */
/*                                                                           */
/*  Return the best supply node to assign to dn.                             */
/*                                                                           */
/*****************************************************************************/

static KHE_SUPPLY_NODE KheFindBestSupplyNode(KHE_RESOURCE_MATCHING_SOLVER rms,
  KHE_DEMAND_NODE dn)
{
  KHE_SUPPLY_NODE sn, best_sn;  int i, c1, c2, c3, c4;
  best_sn = NULL;
  HaArrayForEach(rms->supply_nodes, sn, i)
  {
    /* assign a cost to sn */
    if( KheResourceMatchingEdgeFn(dn, sn, &c1, &c2, &c3) )
    {
      c4 = KheSolnDiversifier(rms->soln) * 5;
      if( sn->resource != NULL )
	c4 += KheResourceResourceTypeIndex(sn->resource) + 1;
      sn->c1 = c1, sn->c2 = c2, sn->c3 = c3, sn->c4 = c4;
    }
    else
      sn->c1 = sn->c2 = sn->c3 = sn->c4 = INT_MAX;

    /* compare sn with best_sn */
    if( best_sn == NULL || KheSupplyNodeLT(sn, best_sn) )
      best_sn = sn;
  }
  return best_sn;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingMakeBestAsst(KHE_RESOURCE_MATCHING_SOLVER rms,   */
/*    KHE_DEMAND_NODE dn)                                                    */
/*                                                                           */
/*  Make the best assignment to dn.                                          */
/*                                                                           */
/*****************************************************************************/

static void KheResourceMatchingMakeBestAsst(KHE_RESOURCE_MATCHING_SOLVER rms,
  KHE_DEMAND_NODE dn)
{
  int i, j;  KHE_SUPPLY_NODE sn;
  KHE_TASK task; KHE_MONITOR m;  KHE_EVENT_RESOURCE er;
  KHE_COST cost;  KHE_MARK mark;

  if( DEBUG4 )
  {
    fprintf(stderr, "[ KheResourceMatchingMakeBestAsst(rms, ");
    KheDemandNodeDebug(dn, 1, -1, stderr);
    fprintf(stderr, "), initial cost %.5f\n",
      KheCostShow(KheSolnCost(rms->soln)));

    /* find the relevant monitors */
    HaArrayForEach(dn->tasks, task, i)
    {
      er = KheTaskEventResource(task);
      if( er != NULL )
	for( j = 0;  j < KheSolnEventResourceMonitorCount(rms->soln, er);  j++ )
	{
	  m = KheSolnEventResourceMonitor(rms->soln, er, j);
	  KheMonitorDebug(m, 2, 2, stderr);
	}
    }
  }

  /* only do anything if assignment is needed */
  if( KheDemandNodeNonAssignmentHasCost(dn) )
  {
    if( dn->grouped )
    {
      /* assign one resource to all tasks, or to none */
      sn = KheFindBestSupplyNode(rms, dn);
      if( sn->resource != NULL )
      {
	cost = KheSolnCost(rms->soln);
        mark = KheMarkBegin(rms->soln);
        KheDemandNodeAssignAllTasks(dn, sn->resource);
	KheMarkEnd(mark, KheSolnCost(rms->soln) >= cost);
      }
    }
    else
    {
      /* one assignment to make for each task; we visit the tasks in */
      /* reverse order so that the first remains free for testing with */
      HaArrayForEachReverse(dn->tasks, task, i)
      {
	sn = KheFindBestSupplyNode(rms, dn);
	if( sn->resource != NULL )
	{
	  cost = KheSolnCost(rms->soln);
	  mark = KheMarkBegin(rms->soln);
          KheDemandNodeAssignOneTask(dn, task, sn->resource);
	  KheMarkEnd(mark, KheSolnCost(rms->soln) >= cost);
	}
      }
    }
  }
  if( DEBUG4 )
    fprintf(stderr,
      "] KheResourceMatchingMakeBestAsst returning, final cost %.5f\n",
      KheCostShow(KheSolnCost(rms->soln)));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceMatchingMakeBestAsst(KHE_RESOURCE_MATCHING_SOLVER rms,   */
/*    KHE_DEMAND_NODE dn)                                                    */
/*                                                                           */
/*  Make the best assignment to dn.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheResourceMatchingMakeBestAsstOld(KHE_RESOURCE_MATCHING_SOLVER rms,
  KHE_DEMAND_NODE dn)
{
  int c1, c2, c3, c4, i, j, ti, si;  KHE_SUPPLY_NODE sn;  bool assigned;
  KHE_TASK task; KHE_MONITOR m;  KHE_EVENT_RESOURCE er;

  if( DEBUG4 )
  {
    fprintf(stderr, "[ KheResourceMatchingMakeBestAsst(rms, ");
    KheDemandNodeDebug(dn, 1, -1, stderr);
    fprintf(stderr, "), initial cost %.5f\n",
      KheCostShow(KheSolnCost(rms->soln)));

    ** find the relevant monitors **
    HaArrayForEach(dn->tasks, task, i)
    {
      er = KheTaskEventResource(task);
      if( er != NULL )
	for( j = 0;  j < KheSolnEventResourceMonitorCount(rms->soln, er);  j++ )
	{
	  m = KheSolnEventResourceMonitor(rms->soln, er, j);
	  KheMonitorDebug(m, 2, 2, stderr);
	}
    }
  }

  ** only do anything if assignment is needed **
  if( KheDemandNodeNonAssignmentHasCost(dn) )
  {
    ** for each supply node sn, store the cost of edge (dn, sn) in sn **
    HaArrayForEach(rms->supply_nodes, sn, i)
      if( KheResourceMatchingEdgeFn(dn, sn, &c1, &c2, &c3) )
      {
	c4 = KheSolnDiversifier(rms->soln) * 5;
	if( sn->resource != NULL )
	  c4 += KheResourceResourceTypeIndex(sn->resource) + 1;
	sn->c1 = c1, sn->c2 = c2, sn->c3 = c3, sn->c4 = c4;
      }
      else
	sn->c1 = sn->c2 = sn->c3 = sn->c4 = INT_MAX;

    ** sort the supply nodes by increasing cost **
    HaArraySort(rms->supply_nodes, &KheSupplyNodeIncreasingCostCmp);
    if( DEBUG4 )
    {
      fprintf(stderr, "  edges sorted by increasing cost:\n");
      HaArrayForEach(rms->supply_nodes, sn, si)
	KheResourceMatchingEdgeDebug(dn, sn, 1, sn->c1, sn->c2, sn->c3,
	  2, stderr);
    }

    ** assign the tasks of dn using the best of them **
    if( DEBUG4 )
      fprintf(stderr, "  best edges:\n");
    if( dn->grouped )
    {
      sn = HaArrayFirst(rms->supply_nodes);
      if( sn->resource != NULL )
        KheDemandNodeAssignAllTasks(dn, sn->resource);
    }
    else
    {
      ti = si = 0;
     while(ti < HaArrayCount(dn->tasks) && si < HaArrayCount(rms->supply_nodes))
      {
	task = HaArray(dn->tasks, ti);
	sn = HaArray(rms->supply_nodes, si);
	if( sn->resource != NULL )
	{
	  ** assign sn->resource to task **
          assigned = KheDemandNodeAssignOneTask(dn, task, sn->resource);
	  if( DEBUG4 )
	  {
	    fprintf(stderr, "  %s: ", assigned ? "assigned" : "not assigned");
            KheResourceMatchingEdgeDebug(dn, sn, 1, sn->c1, sn->c2, sn->c3,
	      0, stderr);
	  }
	  ti++;
	  si++;
	}
	else
	{
	  ** leave all remaining tasks unassigned **
	  if( DEBUG4 )
            KheResourceMatchingEdgeDebug(dn, sn, 1, sn->c1, sn->c2, sn->c3,
	      2, stderr);
	  ti = HaArrayCount(dn->tasks);
	  si = HaArrayCount(rms->supply_nodes);
	}
      }
    }
  }
  else
  {
    if( DEBUG4 )
      fprintf(stderr, "  non-assignment has no cost\n");
  }
  if( DEBUG4 )
    fprintf(stderr,
      "] KheResourceMatchingMakeBestAsst returning, final cost %.5f\n",
      KheCostShow(KheSolnCost(rms->soln)));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheSolnActiveDefectsDebug(KHE_SOLN soln, int verbosity,             */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of defective limit active intervals monitors.                */
/*                                                                           */
/*****************************************************************************/

static void KheSolnActiveDefectsDebug(KHE_SOLN soln, int verbosity,
  int indent, FILE *fp)
{
  KHE_MONITOR m;  int i;
  fprintf(stderr, "%*s[ limit active intervals defects:\n", indent, "");
  for( i = 0;  i < KheSolnDefectCount(soln);  i++ )
  {
    m = KheSolnDefect(soln, i);
    if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
      KheMonitorDebug(m, verbosity, indent + 2, fp);
  }
  fprintf(stderr, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheResourceMatchingSolverSolve(KHE_RESOURCE_MATCHING_SOLVER rms,    */
/*    KHE_RESOURCE_MATCHING_DEMAND_SET rmds, bool matching_off,              */
/*    bool edge_adjust_off, KHE_OPTIONS options)                             */
/*                                                                           */
/*  Do the actual solve.                                                     */
/*                                                                           */
/*****************************************************************************/

bool KheResourceMatchingSolverSolve(KHE_RESOURCE_MATCHING_SOLVER rms,
  KHE_RESOURCE_MATCHING_DEMAND_SET rmds, bool matching_off,
  bool edge_adjust_off, KHE_OPTIONS options)
{
  int i, j, c1, c2, c3, mult, ti;  KHE_TASK task;
  KHE_SOLN soln;  KHE_SUPPLY_NODE sn;  KHE_FRAME_WORKLOAD fw;
  KHE_DEMAND_NODE dn;  KHE_COST init_cost, cost;
  KHE_MMATCH_NODE msn;  KHE_MARK init_mark;  KHE_EVENT_TIMETABLE_MONITOR etm;

  if( DEBUG3 )
    fprintf(stderr, "[ KheResourceMatchingSolverSolve(rms, %s, %s, %s, opts)\n",
      rmds->debug_string != NULL ? rmds->debug_string : "?",
      matching_off ? "matching_off" : "matching_on",
      edge_adjust_off ? "edge_adjust_off" : "edge_adjust_on");

  /* set fields with a well-defined value during a single solve */
  rms->demand_set = rmds;
  rms->options = options;
  rms->common_frame = KheOptionsFrame(options, "gs_common_frame", rms->soln);
  rms->resource_invariant = KheOptionsGetBool(options, "rs_invariant", false);
  rms->matching_off = matching_off;
  rms->edge_adjust_off = edge_adjust_off;

  /* prepare rmds for solving, if not already done */
  if( !rmds->ready_for_solving )
  {
    if( DEBUG3 )
      fprintf(stderr, "  preparting rmds for solving\n");
    KheResourceMatchingDemandSetPrepareForSolving(rmds);
  }

  /* if no assignments to make, return; if there are no demand nodes, */
  /* then the multiplicity of the "absent" supply node is 0, which will */
  /* fail in the mmatch module, so we need this special case */
  if( HaArrayCount(rmds->demand_nodes) == 0 )
  {
    if( DEBUG3 )
      fprintf(stderr, "] KheResourceMatchingSolverSolve returning false"
	" (no demands)\n");
    return false;
  }

  /* boilerplate */
  soln = rms->soln;
  init_cost = KheSolnCost(soln);

  /* add max busy times to resource supply nodes, if required */
  if( !rms->edge_adjust_off && !rms->max_busy_set )
  {
    if( DEBUG3 )
      fprintf(stderr, "  adding max_busy to supply nodes\n");
    etm = (KHE_EVENT_TIMETABLE_MONITOR)
      KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
    fw = KheFrameWorkloadMake(rms->common_frame,
      KheResourceGroupResourceType(rms->resource_group), etm);
    HaArrayForEach(rms->supply_nodes, sn, i)
      if( sn->resource != NULL )
	sn->max_busy = KheFrameResourceMaxBusyTimes(rms->common_frame,
	  fw, sn->resource);
    rms->max_busy_set = true;
    KheFrameWorkloadDelete(fw);
  }

  /* initialize useful fields in the supply nodes */
  if( !rms->edge_adjust_off )
  {
    if( DEBUG3 )
      fprintf(stderr, "  adding curr_busy to supply nodes\n");
    HaArrayForEach(rms->supply_nodes, sn, i)
      if( sn->resource != NULL )
      {
	sn->curr_busy = KheResourceBusyTimes(soln, sn->resource);
	sn->curr_was_used = false;
	/* ***
	sn->curr_at_max = KheResourceAtMaxCount(soln, sn->resource);
	*** */
      }
  }

  /* unassign the tasks in the demand nodes */
  if( DEBUG10 )
  {
    fprintf(stderr, "  KheResourceMatchingSolverSolve initial solution:\n");
    KheSolnDebug(soln, 2, 2, stderr);
    KheSolnActiveDefectsDebug(soln, 2, 2, stderr);
  }
  init_mark = KheMarkBegin(soln);
  HaArrayForEach(rmds->demand_nodes, dn, i)
    KheDemandNodeUnAssignTasks(dn);

  if( rms->matching_off )
  {
    /* without matching */
    if( DEBUG3 )
      fprintf(stderr, "  solving heuristically\n");
    HaArraySort(rmds->demand_nodes, &KheDemandNodeIncreasingDomainSizeCmp);
    HaArrayForEach(rmds->demand_nodes, dn, i)
      KheResourceMatchingMakeBestAsst(rms, dn);
  }
  else
  {
    /* build the mmatch and add supply and demand nodes */
    if( DEBUG3 )
      fprintf(stderr, "  building graph (%d demand nodes, %d supply nodes)\n",
	HaArrayCount(rmds->demand_nodes), HaArrayCount(rms->supply_nodes));
    if( DEBUG7 )
      fprintf(stderr, "    graph supply nodes:\n");
    HaArrayForEach(rms->supply_nodes, sn, i)
    {
      KheSupplyNodeAddMMatchNode(sn, rms->mmatch);
      if( DEBUG7 )
        KheSupplyNodeDebug(sn, 1, 6, stderr);
    }
    if( DEBUG7 )
      fprintf(stderr, "    graph demand nodes:\n");
    HaArrayForEach(rmds->demand_nodes, dn, i)
    {
      KheDemandNodeAddMMatchNode(dn, rms->mmatch);
      if( DEBUG7 )
        KheDemandNodeDebug(dn, 1, 6, stderr);
    }
    if( DEBUG7 )
    {
      c1 = KheHardCost(init_cost);
      c2 = KheSoftCost(init_cost);
      c3 = 0;
      fprintf(stderr, "    graph edges (initial cost ");
      KheResourceMatchingCostDebug(c1, c2, c3, stderr);
      fprintf(stderr, "):\n");
    }
    HaArrayForEach(rmds->demand_nodes, dn, i)
      HaArrayForEach(rms->supply_nodes, sn, j)
        if( KheResourceMatchingEdgeFn(dn, sn, &c1, &c2, &c3) )
	{
	  mult = (sn->resource != NULL ? 1 : KheDemandNodeMultiplicity(dn));
	  KheMMatchAddEdge(dn->mmatch_node, sn->mmatch_node, mult, c1, c2, c3);
	  if( DEBUG7 && !DEBUG9 )
            KheResourceMatchingEdgeDebug(dn, sn, mult, c1, c2, c3, 6, stderr);
	}

    /* solve, extract the matching edges and assign them */
    if( DEBUG3 )
      fprintf(stderr, "  solving the matching graph\n");
    KheMMatchSolve(rms->mmatch);
    if( DEBUG7 )
    {
      fprintf(stderr, "    MMatch after solving:\n");
      KheMMatchDebug(rms->mmatch, 4, 4, stderr);
      fprintf(stderr, "    matching edges:\n");
    }
    if( DEBUG3 )
      fprintf(stderr, "  converting the maximum matching into assignments\n");
    HaArrayForEach(rmds->demand_nodes, dn, i)
      if( dn->grouped )
      {
	/* at most one outgoing edge, defining an assignment to all tasks */
	if( KheMMatchResultEdge(dn->mmatch_node, &msn, &mult, &c1, &c2, &c3) )
	{
	  HnAssert(mult==1, "KheResourceMatchingSolverSolve internal error 1");
	  sn = (KHE_SUPPLY_NODE) KheMMatchSupplyNodeBack(msn);
	  if( DEBUG7 )
	  {
	    KheResourceMatchingEdgeDebug(dn, sn, mult, c1, c2, c3, 6, stderr);
	    HaArrayForEach(dn->tasks, task, j)
	    {
	      fprintf(stderr, "        ");
	      KheTaskDebug(task, 1, -1, stderr);
	      fprintf(stderr, " := ");
	      KheSupplyNodeDebug(sn, 1, 0, stderr);
	    }
	  }
	  if( sn->resource != NULL )
	  {
	    KheDemandNodeAssignAllTasks(dn, sn->resource);
	    sn->curr_was_used = true;
	  }
	}
      }
      else
      {
	/* multiple matching edges */
	ti = 0;
	while( KheMMatchResultEdge(dn->mmatch_node, &msn, &mult, &c1,&c2,&c3) )
	{
	  sn = (KHE_SUPPLY_NODE) KheMMatchSupplyNodeBack(msn);
	  if( DEBUG7 )
	    KheResourceMatchingEdgeDebug(dn, sn, mult, c1, c2, c3, 6, stderr);
	  for( j = 0;  j < mult;  j++ )
	  {
	    HnAssert(ti < HaArrayCount(dn->tasks),
              "KheResourceMatchingSolverSolve internal error 2");
	    task = HaArray(dn->tasks, ti);
	    ti++;
	    if( DEBUG7 )
	    {
	      fprintf(stderr, "        ");
	      KheTaskDebug(task, 1, -1, stderr);
	      fprintf(stderr, " := ");
              KheSupplyNodeDebug(sn, 1, 0, stderr);
	    }
	    if( sn->resource != NULL )
	    {
              KheDemandNodeAssignOneTask(dn, task, sn->resource);
	      sn->curr_was_used = true;
	    }
	  }
	}
	HnAssert(ti == HaArrayCount(dn->tasks),
	  "KheResourceMatchingSolverSolve internal error 3");
      }

    /* clear out the mmatch */
    KheMMatchClear(rms->mmatch);
  }

  /* set up run length for next time */
  if( !rms->edge_adjust_off )
    HaArrayForEach(rms->supply_nodes, sn, i)
      if( sn->resource != NULL )
	sn->curr_run_len = (sn->curr_was_used ? sn->curr_run_len + 1 : 0);

  /* if new solution is not better, return to original solution */
  KheMarkEnd(init_mark, KheSolnCost(soln) >= init_cost);

  /* return true if solution was improved */
  if( DEBUG10 )
  {
    fprintf(stderr, "  KheResourceMatchingSolverSolve final solution:\n");
    KheSolnDebug(soln, 2, 2, stderr);
    KheSolnActiveDefectsDebug(soln, 2, 2, stderr);
  }
  cost = KheSolnCost(soln);
  HnAssert(cost <= init_cost,
    "KheResourceMatchingSolverSolve internal error 4 (init %.5f, final %.5f)",
    KheCostShow(init_cost), KheCostShow(cost));
  if( DEBUG3 )
  {
    if( cost < init_cost )
      fprintf(stderr, "] KheResourceMatchingSolverSolve returning "
	"success: %.5f -> %.5f\n", KheCostShow(init_cost), KheCostShow(cost));
    else
      fprintf(stderr, "] KheResourceMatchingSolverSolve returning "
	"failure: %.5f\n", KheCostShow(cost));
  }
  return KheSolnCost(soln) < init_cost;
}
